/************************************************************************
nborog.c - 'network byte order' implementation of ROG algorithm
 
Copyright (c) 1997 University of Southern California.
All rights reserved.                                            
                                                                
Redistribution and use in source and binary forms are permitted
provided that the above copyright notice and this paragraph are
duplicated in all such forms and that any documentation, advertising
materials, and other materials related to such distribution and use
acknowledge that the software was developed by the University of
Southern California, Information Sciences Institute.  The name of the
University may not be used to endorse or promote products derived from
this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

************************************************************************/

#include <sys/time.h>
#include <sys/resource.h>
#include <sys/kmem_alloc.h>  /* D. Jani - added */
#include <netinet/nborog.h>
#include <netinet/nbomd5.h>

#define BYTE0(x) (x       & 0xff)
#define BYTE1(x) ((x>>8)  & 0xff)
#define BYTE2(x) ((x>>16) & 0xff)
#define BYTE3(x) ((x>>24) & 0xff) 


unsigned long htable [255]; /* local hash table */

/* initializing hash table for the ah */
void NBOinit_htable(){
 int i,j;
 char digest[16];
 struct MD5_CTX context;
 char buffer[32];

 strcpy(buffer,"Hello,thisisfirststringformd5");
 for (i=0; i<255; i++) {
   NBOMD5Init( &context);
   NBOMD5Update(&context, buffer, 16);
   NBOMD5Final(digest, &context);
   htable[i]= (unsigned long) a_to_l(digest);
   for (j=0; j<16;j++)
         buffer[j]=digest[j];
 }
}


static unsigned long counter = 0;  /* counter used for message authentication */

void NBOMD5Init_our (context)
struct MD5_CTX *context;
{
  context->count[0] = context->count[1] = 0;
}


                                  
				
/* macros */
#define make_str(integer, a) {a[3] = '0'+integer % 255; a[2] = '0'+((integer / 255) % 255); a[1] = '0'+((integer / 65025) % 255); a[0] ='0'+ integer / 16581375; a[4] = 0;}


/* This function uses the following algorithm:
   [96 bits of MD5#(ah(counter||string||string_length),counter||key||key_length) ]|| counter
   where counter is a 32 bit word, ah - hasing function, MD5# is a modified MD5 that 
   takes the four 32 bit words produced by ah as the starting values for A,B,C,D
   registers
*/

void NBOhmac_with_counter(context, str, strleng, key, keylen, digest)
struct MD5_CTX *context; 
char *str;
int strleng;
char *key;
int keylen;
char *digest;
{   
   char* buffer1 =  (char *) new_kmem_alloc(sizeof (char) * (20 + keylen), KMEM_NOSLEEP);
   
   unsigned long* ah_buffer=(unsigned long*) kmem_alloc(sizeof(unsigned long)*((strleng/4)+2), KMEM_NOSLEEP);
   char buffer2[5];
   char buffer[32];
   unsigned long hash[4];
   int tmp,i,j;

   counter++; 
   ah_buffer[0]=counter;                            

   /* Another modification - Suggested by J. Touch */

    for(i=0; i<strleng/4;i++)
      ah_buffer[i+1]=((unsigned long *) str)[i];
 
      /* end here */

   ah_buffer[(strleng/4)+1]=(unsigned long) strleng;
   ah(ah_buffer, context->state, (strleng/4+2));      /* producing ah(counter||string||string_length) */

   make_str(counter,buffer1);
   for(i=0; i<keylen;i++)
       buffer1[4+i]=key[i];
   buffer1[4+keylen]=0;
   make_str(keylen,buffer2);
/*   strcat(buffer1,buffer2);      producing counter||key||key_length */
   
   for (i=0;buffer1[i]!='\0';i++);
   for (j=0;buffer2[j]!='\0';j++,i++)
      buffer1[i] = buffer2[j];
   buffer1[i]='\0';

   NBOMD5Init_our( context ); /* initializing md5 with the 4 words of ah */
   
   NBOMD5Update(context, buffer1,keylen+8);
   NBOMD5Final(digest, context);
   make_str(counter,buffer2);
   digest[12]=buffer2[0]; /* appending counter to the digest */
   digest[13]=buffer2[1]; 
   digest[14]=buffer2[2]; 
   digest[15]=buffer2[3]; 
   kmem_free(buffer1, sizeof (char) * (20 + keylen)); 
   kmem_free(ah_buffer, sizeof(unsigned long)*((strleng/4)+2));
  
}


void NBOMD5Init_rog(context)
struct MD5_CTX *context;	/* context */
{
  context->count[0] = context->count[1] = 0;
  /* Load magic initialization constants. */
  context->state[0] = 0;
  context->state[1] = 0;
  context->state[2] = 0;
  context->state[3] = 0;
}




