/************************************************************************
rog.c - This file contains algorithm developed by Dr. Rogaway.
 
Copyright (c) 1997 University of Southern California.
All rights reserved.                                            
                                                                
Redistribution and use in source and binary forms are permitted
provided that the above copyright notice and this paragraph are
duplicated in all such forms and that any documentation, advertising
materials, and other materials related to such distribution and use
acknowledge that the software was developed by the University of
Southern California, Information Sciences Institute.  The name of the
University may not be used to endorse or promote products derived from
this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

************************************************************************/


#include <sys/time.h>
#include <sys/resource.h>
#include <sys/kmem_alloc.h>  /* D. Jani - added */
#include "global.h"
#include <netinet/rog.h>

#define BYTE0(x) (x       & 0xff)
#define BYTE1(x) ((x>>8)  & 0xff)
#define BYTE2(x) ((x>>16) & 0xff)
#define BYTE3(x) ((x>>24) & 0xff) 

/* Convert string to a long integer */

long a_to_l(digest)
char digest[16];
{
int i,j;
long res=0,pow=1;
for(i=0; digest[i]!='\0' && digest[i] >= 48 && digest[i]<=57 ;i++)
    pow *= 10;
pow/=10;
for(j=0;j<i;j++)
{
   if((digest[j]>=48) && (digest[j]<=57))
     {
       res += ((digest[j]-48)*pow);
       pow/=10;
     }  
    else 
        return (res);
}
return (res);
}


unsigned long htable [255]; /* local hash table */

/* initializing hash table for the ah */
void init_htable(){
 int i,j;
 char digest[16];
 struct MD5_CTX context;
 char buffer[32];

 strcpy(buffer,"Hello,thisisfirststringformd5");
 for (i=0; i<255; i++) {
   MD5Init( &context);
   MD5Update(&context, buffer, 16);
   MD5Final(digest, &context);
   htable[i]= (unsigned long) a_to_l(digest);
   for (j=0; j<16;j++)
         buffer[j]=digest[j];
 }

 DAPRINTF(("init-htable():EXIT()\n"));
}


void ah(string, hash, stringlen)
unsigned long* string;
unsigned long* hash;
int stringlen;
{
  unsigned long  temp,A,B,C,D;
  int i, bytetmp=0,tmp;
 
  temp=0;
  A = hash[0];
  B = hash[1];
  C = hash[2];
  D = hash[3];

  DAP(stringlen,d);
  

  for(i=0; i<(stringlen/4); i +=4){
    tmp=A^string[i];

    A ^= htable[BYTE0(tmp)];
    B ^= htable[BYTE1(tmp)];
    C ^= htable[BYTE2(tmp)];
    D ^= htable[BYTE3(tmp)];

    tmp=B^string[i+1];
    B ^= htable[BYTE0(tmp)];
    C ^= htable[BYTE1(tmp)];
    D ^= htable[BYTE2(tmp)];
    A ^= htable[BYTE3(tmp)];
    
    tmp=C^string[i+2];
    C ^= htable[BYTE0(tmp)];
    D ^= htable[BYTE1(tmp)];
    A ^= htable[BYTE2(tmp)];
    B ^= htable[BYTE3(tmp)];
    
    tmp= D^string[i+3];
    D ^= htable[BYTE0(tmp)];
    A ^= htable[BYTE1(tmp)];
    B ^= htable[BYTE2(tmp)];
    C ^= htable[BYTE3(tmp)];
     
  }     
     hash[0] = A;
     hash[1] = B;
     hash[2] = C;
     hash[3] = D; 
    
  /*return  hash;*/
}


static unsigned long counter = 0;  /* counter used for message authentication */

void MD5Init_our (context)
struct MD5_CTX *context;
{
  context->count[0] = context->count[1] = 0;
}


                                  
				
/* macros */
#define make_str(integer, a) {a[3] = '0'+integer % 255; a[2] = '0'+((integer / 255) % 255); a[1] = '0'+((integer / 65025) % 255); a[0] ='0'+ integer / 16581375; a[4] = 0;}


/* This function uses the following algorithm:
   [96 bits of MD5#(ah(counter||string||string_length),counter||key||key_length) ]|| counter
   where counter is a 32 bit word, ah - hasing function, MD5# is a modified MD5 that 
   takes the four 32 bit words produced by ah as the starting values for A,B,C,D
   registers
*/

void hmac_with_counter(context, str, strleng, key, keylen, digest)
struct MD5_CTX *context; 
char *str;
int strleng;
char *key;
int keylen;
char *digest;
{   
   char* buffer1 =  (char *) new_kmem_alloc(sizeof (char) * (20 + keylen), KMEM_NOSLEEP);
   
   unsigned long* ah_buffer=(unsigned long*) new_kmem_alloc(sizeof(unsigned long)*((strleng/4)+2), KMEM_NOSLEEP);
   char buffer2[5];
   char buffer[32];
   unsigned long hash[4];
   int tmp,i,j;

  DAPRINTF(("HMAC_WITH_COUNTER():ENTRY\n"));

   counter++; 
   ah_buffer[0]=counter;                            

   /* D. Jani - BUG ! modification */

    for(i=0; i<strleng/4;i++)
      ah_buffer[i+1]=((unsigned long *) str)[i];
 
      /* end here */
      
   ah_buffer[(strleng/4)+1]=(unsigned long) strleng;
   
   DAPRINTF(("ah():Entry\n"));

   ah(ah_buffer, context->state, (strleng/4+2));      /* producing ah(counter||string||string_length) */

   DAPRINTF(("ah():Exit\n"));
   
   make_str(counter,buffer1);
   for(i=0; i<keylen;i++)
       buffer1[4+i]=key[i];
   buffer1[4+keylen]=0;
   make_str(keylen,buffer2);
   
   for (i=0;buffer1[i]!='\0';i++);
   for (j=0;buffer2[j]!='\0';j++,i++)
      buffer1[i] = buffer2[j];
   buffer1[i]='\0';

   DAPRINTF(("hmac_with_counter():ENTRY\n"));

   MD5Init_our( context ); /* initializing md5 with the 4 words of ah */
   
   MD5Update(context, buffer1,keylen+8);
   MD5Final(digest, context);
   make_str(counter,buffer2);

   DAPRINTF(("\nThe flow comes here..\n"));
  

   digest[12]=buffer2[0]; /* appending counter to the digest */
   digest[13]=buffer2[1]; 
   digest[14]=buffer2[2]; 
   digest[15]=buffer2[3]; 

   DAPRINTF(("\nbuffer attached..\n"));
   kmem_free(buffer1, sizeof (char) * (20 + keylen)); 
   kmem_free(ah_buffer, sizeof(unsigned long)*((strleng/4)+2));

   DAPRINTF(("Freed..\n"));
}


/* Modified md5Init() for use with Rogway's algorithm. Initialize to zero. */

void MD5Init_rog(context)
struct MD5_CTX *context;	/* context */
{
  context->count[0] = context->count[1] = 0;
  /* Load magic initialization constants. */
  context->state[0] = 0;
  context->state[1] = 0;
  context->state[2] = 0;
  context->state[3] = 0;
}




