/************************************************************************
sec.c - This is the main file that contains functions for all algorithms.

Copyright (c) 1997 University of Southern California.
All rights reserved.                                            
                                                                
Redistribution and use in source and binary forms are permitted
provided that the above copyright notice and this paragraph are
duplicated in all such forms and that any documentation, advertising
materials, and other materials related to such distribution and use
acknowledge that the software was developed by the University of
Southern California, Information Sciences Institute.  The name of the
University may not be used to endorse or promote products derived from
this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

************************************************************************/



/* to find trouble points easily, used CHECK, changes, ADD CHANGE */
#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <net/route.h>
#include <sys/protosw.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/in_pcb.h>
#include <netinet/sec.h>
#include <netinet/rog.h>
#include <netinet/nborog.h>
#include "md5.h"
#include "nbomd5.h"
#include <netinet/md5_opt.h>
#include <netinet/nbomd5_opt.h>
#include <netinet/aha.h>
#include <netinet/ip_var.h>
#include <netinet/udp.h>
#include <netinet/udp_var.h>


/* changes */
#define DAA(x) if (!(x)) { DAPRINTF ("Assertion failed! " #x "\n")); /* panic("Assertion failed! " #x "\n"); */}


/* changes - our mbuf structure is different */


/* macros min and max if these have not been defined before */

#ifndef min
#define min(x, y) (x < y) ? x : y
#endif /* min(x, y) */
#ifndef max
#define max(x, y) (x > y) ? x : y
#endif /* max(x, y) */

/* The number of implemented authentication algorithms */
#define NAUTH 9 /* changes */
extern struct protosw inetsw[];
extern u_char ip_protox[IPPROTO_MAX];
extern u_char null_ck;


/* TO ADD - the declarations or m_split and secoutputcheck functions */

/* changes - value 1 for this array */
u_int8 fixedauthkey[1];   

u_int8 fixedcksumkey[1]; /* Not required */ 
u_int32 fixedmd5key[4]; 

int systemsecuritylevel[4];

struct secassoc fixedauth[NAUTH];

static int initted = 0;

 /*changes*/
static void secinit()
{
  u_int8 * charp;
  int l;
  DAPRINTF (("Initializing IP security\n"));

  /* initialize the MD5 key to 0x123456789abcdef (in each byte) */
  charp = (u_int8 *)fixedmd5key;
  for (l = 0; l < 16; l++)
  {   
     charp[l] = (u_int8) l;
  }
  DAPRINTF (("Initialized MD5 key to\n"));
  DADO( hex_print((char *) fixedmd5key, 16) );

  bzero(fixedauth, sizeof(fixedauth)); /* CHECK if size okay */
  
  for (l = 0; l < NAUTH; l ++)
   {
     fixedauth[l].len = sizeof(struct secassoc);
     fixedauth[l].algorithm = l;
     /* Careful here, blast uses 1,2,3 for auth,cksum and md5 resp,
        However, this uses 0, 1, 2 to mean MD5, cksum, and auth  
     */


     switch (l)
     {
      
       case 0: fixedauth[l].spi = htonl(0x40);  /* MD5 */
               fixedauth[l].keylen = sizeof (fixedmd5key);
               fixedauth[l].key =(caddr_t) fixedmd5key;
               null_ck='n';
               break;
       case 1: fixedauth[l].spi= htonl(0x41); /*CKSUM */
               fixedauth[l].keylen = sizeof (fixedcksumkey);
               fixedauth[l].key = (caddr_t) fixedcksumkey;
               null_ck='n';
               break;
       case 2: fixedauth[l].spi=htonl(0x42); /* dummy */
               fixedauth[l].keylen = sizeof(fixedauthkey);
               fixedauth[l].key = (caddr_t)fixedauthkey;
               null_ck='n';
               break;
       case 3: fixedauth[l].spi=htonl(0x43); /* md5-opt*/
               fixedauth[l].keylen = sizeof(fixedmd5key);
               fixedauth[l].key = (caddr_t)fixedmd5key;
               null_ck='n';
               break;
       case 4: fixedauth[l].spi=htonl(0x44); /* aha*/
               fixedauth[l].keylen = sizeof(fixedmd5key);
               fixedauth[l].key = (caddr_t)fixedmd5key;
               break;
       case 5: fixedauth[l].spi=htonl(0x46); /* nbomd5-opt*/
               fixedauth[l].keylen = sizeof(fixedmd5key);
               fixedauth[l].key = (caddr_t)fixedmd5key;
               null_ck='n';
               break;
       case 6: fixedauth[l].spi=htonl(0x48); /* null checksum*/
               fixedauth[l].keylen = sizeof(fixedmd5key);
               fixedauth[l].key = (caddr_t)fixedmd5key;
               null_ck='y';
               break;
       case 7: fixedauth[l].spi=htonl(0x52); /* rogways */
               fixedauth[l].keylen = sizeof(fixedmd5key);
               fixedauth[l].key = (caddr_t)fixedmd5key;
               null_ck='n';
               break;
       case 8: fixedauth[l].spi=htonl(0x54); /* nborogways */
               fixedauth[l].keylen = sizeof(fixedmd5key);
               fixedauth[l].key = (caddr_t)fixedmd5key;
               null_ck='n';
               break;



   
     /* USING THE SAME KEY FOR MD5 MD5opt AHA and NBOMD5opt as of now */

          
    }
   }
  /* changes */				
  bzero(systemsecuritylevel, sizeof(systemsecuritylevel));

  initted++;
}

#define INIT if (!initted) secinit()

/* TO CHECK  - put in MD5 stuff here */

/*
 this *definition* shared by the md5 and md5opt algorithms
*/

struct md5state 
{
  struct MD5_CTX isb;               /* Internal state block for MD5 */
  struct secassoc *assoc;       /* Security association (keys!) */
};

/*----------------------------------------------------------------------
  md5init():   Initialise MD5 state and stash pertinent
               Security Association away in md5state.
----------------------------------------------------------------------*/
static caddr_t md5init(assoc)
     struct secassoc *assoc;
{
  struct md5state *md5state = (struct md5state *) new_kmem_alloc(sizeof(struct md5state), KMEM_NOSLEEP);
  u_int32 zero[4];
  
  DAPRINTF (("md5init():ENTRY\n"));
  
  bzero ( zero, sizeof(zero));

  if (md5state) {
    MD5Init(&(md5state->isb));
    MD5Update(&(md5state->isb), (unsigned char *)assoc->key, assoc->keylen);
    MD5Final(NULL, &(md5state->isb)); 
    md5state->assoc = assoc;
  }
  DAPRINTF(("md5init(): EXIT\n"));
  return (caddr_t)md5state;
  
};


/*----------------------------------------------------------------------
  md5update():  Continue the MD5 calculation over some data in the packet.

----------------------------------------------------------------------*/
static caddr_t md5update(md5state, data, len)
     struct md5state *md5state;
     caddr_t data;
     int len;
{ 
  DAPRINTF (("md5update(): ENTRY\n"));

  MD5Update(&(md5state->isb), (unsigned char *)data, len);

/* CHECK */
#ifdef INET6_DEBUG
  if (ah_debug_count >= 0) {
    int count;
    for (count = 0; count < len; count++, ah_debug_count++) {
      if (!(ah_debug_count % (2 * 8))) {
        printf("\n\t\t\t");
        ah_debug_count = 0;
      }
      if (!(ah_debug_count % 2))
        printf(" ");
      printf("%02x", ((u_int8 *)data)[count]);
    }
  }
#endif /* INET6_DEBUG */

  DAPRINTF(("md5update(): EXIT\n"));
  return (caddr_t)md5state;
};

/*----------------------------------------------------------------------
  md5finish():  Conclude the MD5 calculation for AH and return AH data (?).

----------------------------------------------------------------------*/
static caddr_t md5finish(md5state, data)
     struct md5state *md5state;
     caddr_t data;
{
  u_int32 zero[4];
  int i = (md5state->assoc->keylen % 16);
  
  DAPRINTF(("md5finish(): ENTRY\n"));
  bzero ( zero, sizeof (zero) );
  MD5Update(&(md5state->isb), (unsigned char *)md5state->assoc->key,
            md5state->assoc->keylen);
  MD5Final((unsigned char *)data, &(md5state->isb));
  kmem_free(md5state, sizeof (struct md5state) );
  DAPRINTF(("md5finish(): EXIT\n"));
  return data;
};


/*----------------------------------------------------------------------
               OPTIMISED MD5 ROUTINES BEGIN HERE
------------------------------------------------------------------------*/


/*----------------------------------------------------------------------
  md5optinit():   Initialise MD5 state and stash pertinent
               Security Association away in md5state.
----------------------------------------------------------------------*/
static caddr_t md5opt_init(assoc)
     struct secassoc *assoc;
{
  struct md5state *md5state = (struct md5state *) new_kmem_alloc(sizeof(struct md5state), KMEM_NOSLEEP);
  u_int32 zero[4];
  
  DAPRINTF (("md5opt_init():ENTRY\n"));
  
  bzero ( zero, sizeof(zero));

  if (md5state) {
    MD5optInit(&(md5state->isb));
    MD5optUpdate(&(md5state->isb), (unsigned char *)assoc->key, assoc->keylen);
    MD5optFinal(NULL, &(md5state->isb)); 
    md5state->assoc = assoc;
  }
  DAPRINTF(("md5opt_init(): EXIT\n"));
  return (caddr_t)md5state;
  
};


/*----------------------------------------------------------------------
  md5opt_update():  Continue the MD5 calculation over some data in the packet.

----------------------------------------------------------------------*/
static caddr_t md5opt_update(md5state, data, len)
     struct md5state *md5state;
     caddr_t data;
     int len;
{ 
  DAPRINTF (("md5opt_update(): ENTRY\n"));

  MD5optUpdate(&(md5state->isb), (unsigned char *)data, len);

/* CHECK */
#ifdef INET6_DEBUG
  if (ah_debug_count >= 0) {
    int count;
    for (count = 0; count < len; count++, ah_debug_count++) {
      if (!(ah_debug_count % (2 * 8))) {
        printf("\n\t\t\t");
        ah_debug_count = 0;
      }
      if (!(ah_debug_count % 2))
        printf(" ");
      printf("%02x", ((u_int8 *)data)[count]);
    }
  }
#endif /* INET6_DEBUG */

  DAPRINTF(("md5opt_update(): EXIT\n"));
  return (caddr_t)md5state;
};

/*----------------------------------------------------------------------
  md5opt_finish():  Conclude the MD5 calculation for AH and return AH data (?).

----------------------------------------------------------------------*/
static caddr_t md5opt_finish(md5state, data)
     struct md5state *md5state;
     caddr_t data;
{
  u_int32 zero[4];
  int i = (md5state->assoc->keylen % 16);
  
  DAPRINTF(("md5opt_finish(): ENTRY\n"));
  bzero ( zero, sizeof (zero) );
  MD5optUpdate(&(md5state->isb), (unsigned char *)md5state->assoc->key,
            md5state->assoc->keylen);
  MD5optFinal((unsigned char *)data, &(md5state->isb));
  kmem_free(md5state, sizeof (struct md5state) );
  DAPRINTF(("md5opt_finish(): EXIT\n"));
  return data;
};

/*
_______________________________________________________________________

                     OPTIMIZED MD5 FUNCTIONS END, NBOMD5 optimized
________________________________________________________________________
*/
/*
_______________________________________________________________________

                    NBO  OPTIMIZED MD5 FUNCTIONS BEGIN
________________________________________________________________________
*/

/*----------------------------------------------------------------------
 nbomd5optinit():   Initialise MD5 state and stash pertinent
               Security Association away in md5state.
----------------------------------------------------------------------*/
static caddr_t nbomd5opt_init(assoc)
     struct secassoc *assoc;
{
  struct md5state *md5state = (struct md5state *) new_kmem_alloc(sizeof(struct md5state), KMEM_NOSLEEP);
  u_int32 zero[4];
  
  DAPRINTF (("nbomd5opt_init():ENTRY\n"));
  
  bzero ( zero, sizeof(zero));

  if (md5state) {
    NBOMD5optInit(&(md5state->isb));
    NBOMD5optUpdate(&(md5state->isb), (unsigned char *)assoc->key, assoc->keylen);
    NBOMD5optFinal(NULL, &(md5state->isb)); 
    md5state->assoc = assoc;
  }
  DAPRINTF(("nbomd5opt_init(): EXIT\n"));
  return (caddr_t)md5state;
  
};


/*----------------------------------------------------------------------
  nbomd5opt_update():  Continue the MD5 calculation over some data in the packet.

----------------------------------------------------------------------*/
static caddr_t nbomd5opt_update(md5state, data, len)
     struct md5state *md5state;
     caddr_t data;
     int len;
{ 
  DAPRINTF (("nbomd5opt_update(): ENTRY\n"));

  NBOMD5optUpdate(&(md5state->isb), (unsigned char *)data, len);

/* CHECK */
#ifdef INET6_DEBUG
  if (ah_debug_count >= 0) {
    int count;
    for (count = 0; count < len; count++, ah_debug_count++) {
      if (!(ah_debug_count % (2 * 8))) {
        printf("\n\t\t\t");
        ah_debug_count = 0;
      }
      if (!(ah_debug_count % 2))
        printf(" ");
      printf("%02x", ((u_int8 *)data)[count]);
    }
  }
#endif /* INET6_DEBUG */

  DAPRINTF(("nbomd5opt_update(): EXIT\n"));
  return (caddr_t)md5state;
};

/*----------------------------------------------------------------------
  nbomd5opt_finish():  Conclude the MD5 calculation for AH and return AH data (?).

----------------------------------------------------------------------*/
static caddr_t nbomd5opt_finish(md5state, data)
     struct md5state *md5state;
     caddr_t data;
{
  u_int32 zero[4];
  int i = (md5state->assoc->keylen % 16);
  
  DAPRINTF(("nbomd5opt_finish(): ENTRY\n"));
  bzero ( zero, sizeof (zero) );
 NBOMD5optUpdate(&(md5state->isb), (unsigned char *)md5state->assoc->key,
            md5state->assoc->keylen);
  NBOMD5optFinal((unsigned char *)data, &(md5state->isb));
  kmem_free(md5state, sizeof (struct md5state) );
  DAPRINTF(("nbomd5opt_finish(): EXIT\n"));
  return data;
};

/*
_______________________________________________________________________

               NBO   OPTIMIZED MD5 FUNCTIONS END, AHA start
________________________________________________________________________


/*----------------------------------------------------------------------
  aha_init():   Initialise AHA state and stash pertinent
               Security Association away in md5state.
----------------------------------------------------------------------*/
static caddr_t aha_init(assoc)
     struct secassoc *assoc;
{
  struct md5state *md5state = (struct md5state *) new_kmem_alloc(sizeof(struct md5state), KMEM_NOSLEEP);
  u_int32 zero[4];
  
  DAPRINTF (("aha_init():ENTRY\n"));
  
  bzero ( zero, sizeof(zero));

  if (md5state) {
    AHAInit(&(md5state->isb));
    AHAUpdate(&(md5state->isb), (unsigned char *)assoc->key, assoc->keylen);
    AHAFinal(NULL, &(md5state->isb)); 
    md5state->assoc = assoc;
  }
  DAPRINTF(("aha_init(): EXIT\n"));
  return (caddr_t)md5state;
  
};




/*----------------------------------------------------------------------
  aha_update():  Continue the aha calculation over some data in the packet.

----------------------------------------------------------------------*/
static caddr_t aha_update(md5state, data, len)
     struct md5state *md5state;
     caddr_t data;
     int len;
{ 
  DAPRINTF (("aha_update(): ENTRY\n"));

  AHAUpdate(&(md5state->isb), (unsigned char *)data, len);

/* CHECK */
#ifdef INET6_DEBUG
  if (ah_debug_count >= 0) {
    int count;
    for (count = 0; count < len; count++, ah_debug_count++) {
      if (!(ah_debug_count % (2 * 8))) {
        printf("\n\t\t\t");
        ah_debug_count = 0;
      }
      if (!(ah_debug_count % 2))
        printf(" ");
      printf("%02x", ((u_int8 *)data)[count]);
    }
  }
#endif /* INET6_DEBUG */

  DAPRINTF(("aha_update(): EXIT\n"));
  return (caddr_t)md5state;
};


/*----------------------------------------------------------------------
  aha_finish():  Conclude the AHA calculation for AH and return AH data (?).

----------------------------------------------------------------------*/
static caddr_t aha_finish(md5state, data)
     struct md5state *md5state;
     caddr_t data;
{
  u_int32 zero[4];
  int i = (md5state->assoc->keylen % 16);
  
  DAPRINTF(("aha_finish(): ENTRY\n"));
  bzero ( zero, sizeof (zero) );
  AHAUpdate(&(md5state->isb), (unsigned char *)md5state->assoc->key,
            md5state->assoc->keylen);
  AHAFinal((unsigned char *)data, &(md5state->isb));
  kmem_free(md5state, sizeof (struct md5state) );
  DAPRINTF(("aha_finish(): EXIT\n"));
  return data;
};

/* 
------------------------------------------------------------------
    Start of  of  dummyauth algorithm functions/structures
----------------------------------------------------------------------
*/


struct dummyauthstate 
{
  struct secassoc * assoc;
  /* Given the dummyauth an interface and structures similar to an 
     actual algorithm
  */
};

char dummy;  /*  CHANGE Just a global to get around the malloc error for now */

static caddr_t dummyauthinit ( assoc )
struct secassoc * assoc;
{
  /* this is just a placeholder */

  /*  CHECK, changes is the M_WAIT okay?  */

  /*  struct dummyauthstate * dummyauthstate = (struct dummyauthstate *) malloc	 (   sizeof (struct dummyauthstate), M_TEMP, M_WAIT );
  */
  return (caddr_t) &dummy;


}
static caddr_t dummyauthupdate ( dummyauthstate, data, len )
caddr_t dummyauthstate;
caddr_t data;
int len;

{

  /* No calculation to be done for the dummy case */
  return (caddr_t) dummyauthstate;

}

static  caddr_t dummyauthfinish (dummyauthstate, data)
caddr_t dummyauthstate; /* Actually maybe struct dummyauthstate *  */
caddr_t data;
{
  
 /* As usual, nothing to do */

}
  
/* 
------------------------------------------------------------------
                    End of  dummyauth algorithm functions
----------------------------------------------------------------------
*/

/*
------------------------------------------------------------------
                 Start of checksum() functions
----------------------------------------------------------------------
*/

struct cksumstate
{
  u_int32 sum;
};

/*----------------------------------------------------------------------
  cksuminit():   Initialise the insecure checksum calculations for AH.

----------------------------------------------------------------------*/
static caddr_t cksuminit(assoc)
     struct secassoc *assoc;
{
  struct cksumstate *cksumstate = (struct cksumstate *) new_kmem_alloc(sizeof(struct cksumstate),
                                         KMEM_NOSLEEP);/*changes, CHECK */

  if (cksumstate)
    bzero(cksumstate, sizeof(struct cksumstate));

  return (caddr_t)cksumstate; /* bacs : returned as a character pointer */

};

/*
----------------------------------------------------------------------
  cksumupdate():  Update the insecure checksum calculation with new data.

----------------------------------------------------------------------
*/
static caddr_t cksumupdate(cksumstate, data, len)
     struct cksumstate *cksumstate;
     caddr_t data;
     int len;
{
  u_int32 sum = cksumstate->sum;
  DAPRINTF (("----------------BEGIN-----------------\n"));
  DAPRINTF (("In the cksumupdate() function,\n"));
  DAPRINTF (("State passed in\n"));
  DADO (hex_print ((u_int8 *) cksumstate, 4));
  DAPRINTF (("Data passed in for summing\n"));
  DADO (hex_print (data,len));
  DAPRINTF (("\n\n"));

/* J. Touch - 12/96
   rewrote checksum to be native 32-bit.
*/


/* the goal isn't a perfect checksum, just a data-touching
   emulation. so don't fret about alignment errors
   just force the alignment
*/

  (int)data &= 0xFFFFFFFC;

/* run through the data */

  while ((len -= 4) > 1)
  {
    sum += *(((u_int32 *)data)++);
  }

/* don't worry about how this is folded over; just fudge it */
  sum = (sum & 0xffff) + (sum >>16);
  cksumstate->sum = (sum & 0xffff) + (sum >> 16); /* for last carry */

/* end Joe's mods */

  DAPRINTF (("State passed out\n"));
  DADO( hex_print ( (caddr_t) cksumstate, 4));
  DAP (cksumstate->sum,4x);
  DAPRINTF (("---------------END----------------"));
  return (caddr_t)cksumstate;
}
/*----------------------------------------------------------------------
  cksumfinish():  Conclude the insecure checksum calculation and return
                  the calculated Authentication data.
----------------------------------------------------------------------*/
static caddr_t cksumfinish(cksumstate, data)
     struct cksumstate *cksumstate;
     caddr_t data;
{
  /**((u_int16 *)data) = cksumstate->sum; Just return  zero*/
  *((u_int16 *)data) = (u_int16) 0;
  kmem_free(cksumstate, sizeof(struct cksumstate) );
  return data;
};

/*----------------------------------------------------------------------
                 End of    checksum functions
----------------------------------------------------------------------*/

/*
------------------------------------------------------------------
                 Start of nullchecksum() functions
----------------------------------------------------------------------
*/

struct nullcksumstate
{
  u_int32 sum;
};

/*----------------------------------------------------------------------
  nullcksuminit():   Initialise the insecure checksum calculations for AH.

----------------------------------------------------------------------*/
static caddr_t nullcksuminit(assoc)
     struct secassoc *assoc;
{
  struct nullcksumstate *nullcksumstate = (struct nullcksumstate *) new_kmem_alloc(sizeof(struct nullcksumstate),
                                         KMEM_NOSLEEP);/*changes, CHECK */

  if (nullcksumstate)
    bzero(nullcksumstate, sizeof(struct nullcksumstate));

  return (caddr_t)nullcksumstate; /* bacs : returned as a character pointer */

};

/*
----------------------------------------------------------------------
  nullcksumupdate():  Update the insecure checksum calculation with new data.

----------------------------------------------------------------------
*/
static caddr_t nullcksumupdate(nullcksumstate, data, len)
     struct nullcksumstate *nullcksumstate;
     caddr_t data;
     int len;
{
  u_int32 sum = nullcksumstate->sum;
  DAPRINTF (("----------------BEGIN-----------------\n"));
  DAPRINTF (("In the nullcksumupdate() function,\n"));
  DAPRINTF (("State passed in\n"));
  DADO (hex_print ((u_int8 *) nullcksumstate, 4));
  DAPRINTF (("Data passed in for summing\n"));
  DADO (hex_print (data,len));
  DAPRINTF (("\n\n"));

/* J. Touch - 12/96
   rewrote checksum to be native 32-bit.
*/


/* the goal isn't a perfect checksum, just a data-touching
   emulation. so don't fret about alignment errors
   just force the alignment
*/

  (int)data &= 0xFFFFFFFC;

/* run through the data */

  while ((len -= 4) > 1)
  {
    sum += *(((u_int32 *)data)++);
  }

/* don't worry about how this is folded over; just fudge it */
  sum = (sum & 0xffff) + (sum >>16);
  nullcksumstate->sum = (sum & 0xffff) + (sum >> 16); /* for last carry */

/* end Joe's mods */

  DAPRINTF (("State passed out\n"));
  DADO( hex_print ( (caddr_t) nullcksumstate, 4));
  DAP (nullcksumstate->sum,4x);
  DAPRINTF (("---------------END----------------"));
  return (caddr_t)nullcksumstate;
}
/*----------------------------------------------------------------------
  nullcksumfinish():  Conclude the insecure checksum calculation and return
                  the calculated Authentication data.
----------------------------------------------------------------------*/
static caddr_t nullcksumfinish(nullcksumstate, data)
     struct nullcksumstate *nullcksumstate;
     caddr_t data;
{
  /**((u_int16 *)data) = nullcksumstate->sum; Just return  zero*/
  *((u_int16 *)data) = (u_int16) 0;
  kmem_free(nullcksumstate, sizeof(struct nullcksumstate) );
  return data;
};

/*----------------------------------------------------------------------
                 End of    nullcksum functions
----------------------------------------------------------------------*/

/*----------------------------------------------------------------------
       Rogaways algorithm 
-----------------------------------------------------------------*/

struct rogstate 
{
  struct MD5_CTX isb;               /* Internal state block for MD5 */
  struct secassoc *assoc;       /* Security association (keys!) */
  caddr_t digest;
};


/*----------------------------------------------------------------------
  roginit():   Initialise MD5 state and stash pertinent
               Security Association away in md5state.
----------------------------------------------------------------------*/
static caddr_t roginit(assoc)
     struct secassoc *assoc;
{
  struct rogstate *rogstate = (struct rogstate *) new_kmem_alloc(sizeof(struct rogstate), KMEM_NOSLEEP);

  u_int32 zero[4];

  rogstate->digest = (caddr_t) new_kmem_alloc(16 * sizeof(unsigned char), KMEM_NOSLEEP);   /* D. Jani - added */

  DAPRINTF (("roginit():ENTRY\n"));
  
  bzero ( zero, sizeof(zero));


  if (rogstate) {
    init_htable();
    MD5Init_rog(&(rogstate->isb));
    rogstate->assoc = assoc;
  }
  DAPRINTF(("roginit(): EXIT\n"));
  return (caddr_t)rogstate;
  
};

/*----------------------------------------------------------------------
  rogupdate():  Continue ah_mac data calculation over some data in the packet.

----------------------------------------------------------------------*/
static caddr_t rogupdate(rogstate, data, len)
     struct rogstate *rogstate;
     caddr_t data;
     int len;
{
  DAPRINTF (("rogupdate(): ENTRY\n"));

  MD5Update(&(rogstate->isb), (unsigned char *)data, len);
  hmac_with_counter(&(rogstate->isb), data, len, rogstate->assoc->key, rogstate->assoc->keylen, rogstate->digest); /* D. Jani - changed from digest */    

/* CHECK */
#ifdef INET6_DEBUG
  if (ah_debug_count >= 0) {
    int count;
    for (count = 0; count < len; count++, ah_debug_count++) {
      if (!(ah_debug_count % (2 * 8))) {
        printf("\n\t\t\t");
        ah_debug_count = 0;
      }
      if (!(ah_debug_count % 2))
        printf(" ");
      printf("%02x", ((u_int8 *)data)[count]);
    }
  }
#endif /* INET6_DEBUG */

  DAPRINTF(("rogupdate(): EXIT\n"));
  return (caddr_t)rogstate;
};


/*----------------------------------------------------------------------
  rogfinish():  Conclude ah_mac data calculation for AH and return AH data (?).

----------------------------------------------------------------------*/
static caddr_t rogfinish(rogstate, data)
     struct rogstate *rogstate;
     caddr_t data;
{
  u_int32 zero[4];

  data = (caddr_t) rogstate->digest;
  DAPRINTF(("rogfinish(): ENTRY\n"));
  bzero ( zero, sizeof (zero) );

  kmem_free(rogstate->digest, 16 * sizeof(unsigned char)); /* D. Jani */
  kmem_free(rogstate, sizeof (struct rogstate) );
  DAPRINTF(("rogfinish(): EXIT\n"));
  return data ;
};

/*--------------------------------------------------------------------

                           NBO ROGAWAYS 

-----------------------------------------------------------------------*/


/*----------------------------------------------------------------------
  nbo_roginit():   Initialise MD5 state and stash pertinent
               Security Association away in md5state.
----------------------------------------------------------------------*/

static caddr_t nbo_roginit(assoc)
     struct secassoc *assoc;
{
  struct rogstate *rogstate = (struct rogstate *) new_kmem_alloc(sizeof(struct rogstate), KMEM_NOSLEEP);
  u_int32 zero[4];

  rogstate->digest = (caddr_t) new_kmem_alloc(16 * sizeof(unsigned char), KMEM_NOSLEEP);   /* D. Jani - added */
  
  DAPRINTF (("roginit():ENTRY\n"));
  
  bzero ( zero, sizeof(zero));


  if (rogstate) {
    NBOinit_htable();
    NBOMD5Init_rog(&(rogstate->isb));
    rogstate->assoc = assoc;
  }
  DAPRINTF(("roginit(): EXIT\n"));
  return (caddr_t)rogstate;
  
};

/*----------------------------------------------------------------------
  nbo_rogupdate():  Continue ah_mac data calculation over some data in the packet.

----------------------------------------------------------------------*/

static caddr_t nbo_rogupdate(rogstate, data, len)
     struct rogstate *rogstate;
     caddr_t data;
     int len;
{

  NBOMD5Update(&(rogstate->isb), (unsigned char *)data, len);
  NBOhmac_with_counter(&(rogstate->isb), data, len, rogstate->assoc->key, rogstate->assoc->keylen, rogstate->digest);    

/* CHECK */
#ifdef INET6_DEBUG
  if (ah_debug_count >= 0) {
    int count;
    for (count = 0; count < len; count++, ah_debug_count++) {
      if (!(ah_debug_count % (2 * 8))) {
        printf("\n\t\t\t");
        ah_debug_count = 0;
      }
      if (!(ah_debug_count % 2))
        printf(" ");
      printf("%02x", ((u_int8 *)data)[count]);
    }
  }
#endif /* INET6_DEBUG */

  DAPRINTF(("rogupdate(): EXIT\n"));
  return (caddr_t)rogstate;
};


/*----------------------------------------------------------------------
  nbo_rogfinish():  Conclude ah_mac data calculation for AH and return AH data (?).

----------------------------------------------------------------------*/
static caddr_t nbo_rogfinish(rogstate, data)
     struct rogstate *rogstate;
     caddr_t data;
{
  u_int32 zero[4];

  data = (caddr_t) rogstate->digest;
  DAPRINTF(("rogfinish(): ENTRY\n"));
  bzero ( zero, sizeof (zero) );

  kmem_free(rogstate->digest, 16 * sizeof(unsigned char));
  kmem_free(rogstate, sizeof (struct rogstate) );
  DAPRINTF(("rogfinish(): EXIT\n"));
  return data ;
};


/*-----------------------------------------------------------------*/




/* TO CHECK , changes*/

/*  arguments commented out for cc */

struct auth_algorithm 
{
  caddr_t (*init)(/*struct secassoc * */);      /* ptr to init function */
  caddr_t (*update)(/* caddr_t, caddr_t, int */); /* ptr to update function */
  caddr_t (*finish)( /* caddr_t, caddr_t */);     /* ptr to finish function */
  int datalen;                               /* Length (in u_int8s) of output 
					       data. MUST be a multiple of 4 */
}auth_algorithms[]={ 
{ md5init, md5update, md5finish, 16 },    /* algorithm 0*/
{cksuminit, cksumupdate, cksumfinish, 4 },/* algorithm 1*/
{dummyauthinit, dummyauthupdate, dummyauthfinish, 16}, /* algorithm 2 */
/* algorithm no. two, just the authentication header carried, no auth */
{ md5opt_init, md5opt_update, md5opt_finish, 16 },    /* algorithm 3 */
/* Optimized md5 code */
{ aha_init, aha_update, aha_finish, 16},   /* algorithm 4 */
{ nbomd5opt_init, nbomd5opt_update, nbomd5opt_finish, 16 },    /* algorithm 5*/
{nullcksuminit, nullcksumupdate, nullcksumfinish, 4 },  /* algorithm 6 */
{roginit, rogupdate, rogfinish, 16}, /* algorithm 7 */
{nbo_roginit, nbo_rogupdate, nbo_rogfinish, 16}  /* algorithm 8 */
};

/* TO ADD  initialize the different algorithms */


/* 
----------------------------------------------------------------------
Routine:     secahcalc()
Output :     Gives the result of the calculation of AH data over an IP packet
    
----------------------------------------------------------------------
*/

int secahcalc ( packet, assoc, algorithm, data)
struct mbuf * packet;
struct secassoc * assoc;
struct auth_algorithm * algorithm;
char * data;
{
/* Taken from the NRL release */

  u_int8 type;
  struct mbuf *m;
  int i;
  int j;
  int rval = -1;
  unsigned int option_length;
  caddr_t state;
  int indexoff = 1;
  int authbytes; /* No of bytes of the header and auth data in the mbuf */
  struct ip *ip;
  struct in_addr savedst;
  u_int8  savetos;
  u_int8  savettl;
  u_int16 savesum;
  u_int16 saveoff;
  u_int8 *optp;
  u_int8 *optend;
  u_int8 zero[32];
  struct auth * ah; 

  DAPRINTF (("Entering secahcalc()\n"));
  
/*  
    DPRINTF(IDL_GROSS_EVENT, ("Entering ipsec_ah_calc\n"));
    DDO(IDL_GROSS_EVENT, dump_tcpdump(packet));
    DPRINTF(IDL_GROSS_EVENT, ("..."));
*/

  if (!packet->m_len)
    packet = m_free(packet);

  type = (*(mtod(packet, u_int8 *))) >> 4; /* CHANGE */

  if (!(state = algorithm->init(assoc)))
  {
    DAPRINTF (("AH algorithm init failed! \n"));
    /*DPRINTF(IDL_CRITICAL, ("AH algorithm init failed!\n")); */

    /* changes CHECK by making the kernel fail here  lots of times and see
       if there is an mbuf leak. 
    */
    /*
       Point is, I don't think an m_freem() is required here since if a NULL 
       pointer is going to be returned, the routine which called secahcalc()
       (ahinput, ahoutput) is responsible for releasing the space in case of an
        error.
       Why not just use a goto ahcret
    */
    /* m_freem(packet); */
    return NULL;
  }

    /* changes - Removed ipv6 stuff, moved declarations to top */
    /* Must be IPv4; If this is ST-2, we'll die badly. */
  
  
    /* 
     create a 32-byte array containing all zeroes;
     this will be used whenever we need to zero an option for AH calc. 
    */
    bzero(zero, sizeof(zero));

    /*DDO(IDL_GROSS_EVENT, ah_debug_count = 0); changes */
    
    ip = mtod(packet, struct ip *);
    /* bacs: save all the fields which are not being authenticated and
       hence have to be zero'd for the calculation 
    */

    bcopy(&(ip->ip_dst), &savedst, sizeof(struct in_addr));

    type = ip->ip_p; /* CHECK */

    savetos = ip->ip_tos;
    savettl = ip->ip_ttl;
    savesum = ip->ip_sum;
    saveoff = ip->ip_off;

    ip->ip_tos = 0;
    ip->ip_ttl = 0;
    ip->ip_sum = 0;
    ip->ip_off = 0;
     /*
        CHECK: WHY WAS THE DESTINATION FIELD NOT ZEROED OUT HERE BEFORE
        THE CALCULATION WAS DONE: CHANGE
    */

    if (!(state = algorithm->update(state, (caddr_t)ip,
                                    sizeof(struct ip)))) 
    {
       DAPRINTF (("secahcalc:algorithm->update failed\n"));      
      /*DPRINTF(IDL_CRITICAL, ("algorithm->update failed (IPv4 header)\n"));*/
      goto ahcret;
    }
    
    /* Restoring all the previously zeroed out fields in the ip header */
    ip->ip_tos = savetos;
    ip->ip_ttl = savettl;
    ip->ip_sum = savesum;
    ip->ip_off = saveoff;
 
    bcopy(&savedst, &(ip->ip_dst), sizeof(struct in_addr));

    /* Make optp point to the first byte of ip options */
    optp = (u_int8 *)ip + sizeof(struct ip);

    /* make optend point to the first byte (in mem) after the IP options */
    optend = (u_int8 *)ip + (ip->ip_hl << 2);
    

  /* pass 2: Call (algorithm->update)() to calculate AH value,
           zeroing where appropriate */

    while(optp < optend) {
      
      DAP (optp,8x); 
      DAP (optend,8x);     
      /*DP(GROSS_EVENT, optp, 08x);
      /*DP(GROSS_EVENT, optend, 08x);*/
      switch(*optp)
      {

        /* ZEROED for AH computation */
        case IPOPT_EOL:
          if (!(state = algorithm->update(state, (caddr_t)zero, 1))) 
          {
            DAPRINTF (("algorithm->update failed (EOL) \n"));
            /*DPRINTF(IDL_CRITICAL, ("algorithm->update failed (EOL)\n"));*/
            goto ahcret;
          }
          optp = optend;
          break;

        case IPOPT_NOP:
          if (!(state = algorithm->update(state, (caddr_t)zero, 1))) 
          {
              DAPRINTF (("algorithm->update failed (NOP) \n"));
            /*DPRINTF(IDL_CRITICAL, ("algorithm->update failed (NOP)\n"));*/
            goto ahcret;
          }
          optp++;
          break;

        case IPOPT_RR:
        case IPOPT_TS:
          /* While RFC-791 is not crystal clear about the above two options,
             there is good reason to believe that -- AS IMPLEMENTED in
             The Internet -- these two options vary during transit in
             a manner unpredictable at the sending side.
          */
        case IPOPT_LSRR:
        case IPOPT_SSRR:
          /*  The above two vary unpredictably
             because of the record-route action. */
 
        case 10:    /* Z Su's "Experimental Measurement" */
        case 15:    /* VerSteeg's ENCODE option */
        case 142:   /* Estrin's VISA option */
        case 144:   /* Lee's IMI Traffic Descriptor */
        case 145:   /* EIP per RFC-1385 */
        case 205:   /* Greg Finn's "Experimental Flow Control" */

          /* This last set of  option numbers are from the
             Assigned Numbers RFC; included only for completeness.
          */
          {
            option_length = optp[1];
            if (!(state = algorithm->update(state, (caddr_t)zero,
                                            option_length))) 
            {
               DAPRINTF(("algorithm->update failed (TS/RR) \n"));
               /* DPRINTF(IDL_CRITICAL, ("algorithm->update failed (TS/RR)\n"));*/
               goto ahcret;
            }
            optp += option_length;
          }
          break;


	case 11:  /* MTUP   - MTU Probe */
        case 12:  /* MTUR   - MTU Reply */
                  /* These have a length of 4 bytes */
          {
            if (!(state = algorithm->update(state, (caddr_t)zero, 4)))
            {
              DAPRINTF (("algorithm->update failed (MTU)\n"));
              /*DPRINTF(IDL_CRITICAL, ("algorithm->update failed (MTU)\n"));*/
              goto ahcret;
            }
            optp += 4;
          }
          break;
        case 82:  /* TR     - Traceroute; has fixed length of 12 bytes */
          {
            if (!(state = algorithm->update(state, (caddr_t)zero, 12))) 
            { 
              DAPRINTF (("algorithm->update failed (tracerte)\n"));
              /* DPRINTF(IDL_CRITICAL, ("algorithm->update failed (tracerte)\n"));*/
              goto ahcret;
            }
            optp += 12;
          }
          break;


        /* INCLUDED in AH computation */
        /* ------------------------------------------------------------------ */ 
        case IPOPT_SATID:
        case IPOPT_SECURITY:  /* IPSO - BSO, see RFC-1108 */
        case 133:             /* IPSO - ESO, see RFC-1108 */
        case 134:             /* CIPSO, no published RFC */
          option_length = optp[1];
          if (!(state = algorithm->update(state, (caddr_t)optp,
                                          option_length))) 
          {
            DLOG ((LOG_DEBUG,"algorithm->update failed (opt %d)\n",*optp)); 
            /* DPRINTF(IDL_CRITICAL, ("algorithm->update failed (opt %d)\n",
                                   *optp));*/
            goto ahcret;
          }
          optp += option_length;
          break;

        /* anything else we assume is completely invariant */
        default:
          if (!(state = algorithm->update(state, (caddr_t)optp, optp[2])))
          {
            DAPRINTF (("algorithm->update failed (default)\n"));
            /*DPRINTF(IDL_CRITICAL, ("algorithm->update failed (default)\n"));*/
            goto ahcret;
          } /* end if */

          optp += optp[1];
      } /* end switch (*optp) */
   } /* end while (optp < voptend) */


   DAPRINTF (("AH calc: Starting main loop\n"));

   /* DPRINTF(IDL_EVENT, ("AH calc: Starting main loop\n"));*/
   /* CHANGE if no problem in debug stage */
#ifdef AH_DEBUG
   if(packet->m_len > (ip->ip_hl<< 2)  )
   {
      DAPRINTF (("secahcalc():ERROR: Packet header not preparsed????\n"));
   }
#endif AH_DEBUG


   m = packet->m_next;
   if(m) 
   {
        
     /*  DP(GROSS_EVENT, m, 08x);
         DP(GROSS_EVENT, type, d);
     */
    DAP(type,d); 
    if (type == 51)   /* NUMBER for Authentication headers */
    {     char zero[32]; /* CHECK scope */
          ah = mtod(m, struct auth *);
          
          DAPRINTF (("secahcalc() : Now summing AH\n"));
          /* DPRINTF(IDL_GROSS_EVENT, ("AH calc: Now summing AH\n"));*/

         if (!(state = algorithm->update(state, mtod(m, caddr_t), sizeof(struct auth))))
         {  
            DAPRINTF (("secahcalc:auth header update failed\n"));
            /* DPRINTF(IDL_CRITICAL, ("AH calc: auth header update failed\n"));*/
            goto ahcret;
         }
         type = ah->auth_nexthdr;
         
        /* DP(GROSS_EVENT, type, d);*/

        bzero(zero, sizeof(zero)); /* ????*/
        i = ah->auth_datalen << 2;
        /*DP(GROSS_EVENT, i, d);*/
        DAP (ah->auth_datalen,d);

        /* Perform the update with a block of zero's as big as the auth data*/
        while (i > 0) 
        {
           j = min(sizeof(zero), i);
           if (!(state = algorithm->update(state, zero, j))) 
           {
             DAPRINTF(("AH calc: (auth data) update failed"));
             /*DPRINTF(IDL_CRITICAL, ("AH calc: (auth data) update failed\n"));*/
             goto ahcret;
           }
           i -= j;
        };
          
        authbytes = sizeof (struct auth) + (ah->auth_datalen << 2);/*CHECK */
        DAP( authbytes,d);
        DAP (type,d);
        switch (type)
        {
             /* type should be containing the protocol the IP is carrying - */
  
            case IPPROTO_IP:
                 DAPRINTF (("secahcalc():IP within IP (tunnel)?\n"));
            case IPPROTO_ICMP:
            case IPPROTO_IGMP:
            case IPPROTO_TCP:
            case IPPROTO_UDP:
         
            DAPRINTF (("secahcalc:Going to make transport update\n"));
           if ( m->m_len - authbytes)
           {  
              DAPRINTF (("m->m_len -authbytes is non zero, summing this first\n"));
              if (!(state = algorithm->update(state,((char *) ah) + authbytes, m->m_len - authbytes )))
             {
                DAPRINTF(("AH calc: (auth extra data) update failed\n"));
                /*DPRINTF(IDL_CRITICAL, ("AH calc: (auth data) update failed\n"));*/
                goto ahcret;
             }
           }
           m = m->m_next;
       
           /* rval = 0; changes, this should be done after update?? CHECK */
           while (m) 
           {
             DAPRINTF (("In the main loop now....\n"));
	     DADO ( dump_chain(m, DETAIL));  /* D. Jani added */

             if (!(state = algorithm->update(state, mtod(m, caddr_t),
                                           m->m_len))) 
             {
               DAPRINTF(("AH calc: (transport) update failed\n"));
               /*DPRINTF(IDL_CRITICAL, ("AH calc: (transport) update failed\n"));*/
               goto ahcret;
             }

	     DAPRINTF(("UPDATE is done...\n"));
             m = m->m_next;
	     DAPRINTF(("Next mbuf...\n"));
           }
           rval = 0; /* Everything okay since transport update done successfully*/        /* changes  CHECK */  
           break;         
           default:
               DAPRINTF(("What do I do with this protocol???\n"));         
       }
   

     }else
     {      
       DAPRINTF (("secahcalc: illegal packet format\n"));  
       /*DPRINTF(IDL_CRITICAL, ("AH calc: illegal packet format\n"));*/
       goto ahcret;
     }
  }; /* end if(m) */
 /* intentionally fall thru the ahcret code */

 ahcret:
  /* cleanup then bail */
  /*  DPRINTF(IDL_MAJOR_EVENT, ("\nipsec_ah_calc() finishing. rval=%d\n", rval));*/

  DAPRINTF (("secahcalc() finishing," ));
  DAP (rval,d);
  algorithm->finish(state, data);
#ifdef AH_DEBUG
  if( (rval== 0) && (ah != (struct auth *) 0) )
  {  DAPRINTF (("The data calculated is "));
     DADO ( hex_print ( (char *) (data),(ah->auth_datalen << 2) ) );
     DAPRINTF (("The data in the packet (IF INPUT SIDE) is "));
     DADO ( hex_print ( (char * ) (ah + 1) , (ah->auth_datalen << 2) ) );
  }
#endif AH_DEBUG

  return rval;
}




/*
----------------------------------------------------------------------
 Routine:     hex_print ()
 Funtion:     print out a series of hex numbers

----------------------------------------------------------------------
*/

int hex_print ( cp, num)
u_char * cp;
int num;
{
  int i;
  DAPRINTF (("\nhex_print() \n"));
  DLOG (( LOG_DEBUG,"Number of bytes is %d\n",num));

  for (i = 0; i < num; i++)
  { 
    DLOG ((LOG_DEBUG,"%2x",cp[i]));
  }  
  DAPRINTF (("\n")); 

}
/* Define DETAIL to print info on each of the mbuf in chain */

/*
------------------------------------------------------------------------
Routine:     dump_chain()
Funtion:     total bytes in chain; gives bytes per mbuf if detail = 1;
Output;      int no. of bytes
    
----------------------------------------------------------------------
*/
int dump_chain(m,detail)
struct mbuf * m;
int detail;
{
    int i = 1;
    int sum = 0	;
    while(m != (struct mbuf *)0)
    {
       if(detail)
       {
          if (i==1)
          {
            i++;
            DAPRINTF (("MBUF CHAIN\n"));        
           
          }else
          {
              DAPRINTF (("Next mbuf\n"));
          }
  
          DAP (m->m_len,d);
          DAP (m->m_off,d);
         
       }
         
          sum += m->m_len;
          m = m->m_next;
          
    }
    if (detail)
    {
      DLOG((LOG_DEBUG,"TOTAL bytes in the chain = %d\n",sum));  
      
    }
   
    return(sum);
  
}
/* 
----------------------------------------------------------------------
Routine:     print_ip_addr
Funtion:     printout an ip address 
To do : See if conversion is required bet network and host byte order    
----------------------------------------------------------------------
*/
void
print_ip_addr(data)
     char *data;
{
   int i;

   for (i = 0; i < IP_ADDR_LEN; i++)
      DLOG((LOG_DEBUG, "%s%d", (i)?".":" ", (u_char)data[i]));
      DAPRINTF (("\n")); /* newline after the address */
}
                 
/* 
----------------------------------------------------------------------
Routine:     ip_header
Funtion:     printout all fields of the IP header 
To do : Print out the host addresses also; id, checksum may be screwed up  
----------------------------------------------------------------------
*/
int ip_header ( m,byte_order )
struct mbuf * m;
int byte_order;  /* Network or host order */
{
   struct ip * ip;
   u_char hlen;
   u_char version;
   u_char tos;
   u_short total_length, t_total_length; /* should this be a short?? */
   u_short identity, t_identity;
   u_short offset, t_offset;
   u_short more_fragments;
   u_short dont_fragment;
   u_char ttl;
   u_char protocol;
   u_short checksum;
   

   ip = mtod (m, struct ip *);
   hlen = ip->ip_hl<< 2;
   version = ip->ip_v ;
   tos = ip->ip_tos;
   if ( byte_order = HOST_ORDER)
   {
     t_identity = ip->ip_id;
     t_offset = ip->ip_off;
     t_total_length = ip->ip_len;
    
   }else
   {
     t_identity = ntohs (ip->ip_id); 
     t_offset = ntohs (ip->ip_off);
     t_total_length = ntohs (ip->ip_len);
   }




    total_length = t_total_length;
    identity = t_identity;
    offset = (t_offset & ~ (IP_DF | IP_MF) )<< 3;
    more_fragments = t_offset & IP_MF;
    dont_fragment = t_offset & IP_DF;
    ttl = ip->ip_ttl;
    protocol = ip->ip_p;
    
    DAP (hlen,d);
    DAP (version,d);
    DAP (tos,d);
    DAP (total_length,d);    
    DAP (identity,d);
    DAP (offset,d);
    if (more_fragments)
    {
       DAPRINTF(("MORE fragments set\n"));
    }
    if (dont_fragment)
    {
       DAPRINTF(("DONT fragment flag set\n"));
    }
    DAP (ttl,d);
    DAP (protocol,d);
    DAP ( ntohs(ip->ip_sum),x);
    DAPRINTF (("Source of packet is\n")); 
     print_ip_addr ( (char *)	& (ip->ip_src) );
    DAPRINTF (("Destination of packet is\n"));
    print_ip_addr ( (char *) & (ip->ip_dst) );


}
 




/* 
----------------------------------------------------------------------
Routine:     ahoutput
Funtion:     Construct the AH, arrange mbufs, call secahcalc 
Output:      Complete packet with the AH in it  
    
----------------------------------------------------------------------
*/



struct mbuf * ahoutput(outgoing,flags)
     struct mbuf *outgoing;
     u_int16 flags;
{
  struct secassoc *assoc;
  struct mbuf *ahm , *m;
  struct auth *ah;
  struct auth_algorithm *algorithm;
  /* changes : Might not be needed */
  /*  struct in6_hdrindex index[IPV6_MAXHDRCNT];  */
  u_char nexthdr;   /* Transport protocol for the packet */
  int iphlen; /*To store the length of the ip header including options*/
  struct mbuf * n ; 
  struct mbuf * n0 = (struct mbuf *) 0;  /* CHANGE  free mbufs properly */
  struct mbuf * ahm0 = (struct mbuf *) 0;

  struct ip *ip;

  INIT;
    
/*changes We dont come to this function at all if we dont need authenticated 
packets */
/*  if (!(assoc = in6_secoutputcheck(outgoing, SS_AUTHENTICATION)))
    return outgoing;
*/

DAPRINTF ( ("Entered ahoutput\n"));
DADO ( ip_header(outgoing, HOST_ORDER) );
DADO ( dump_chain (outgoing, DETAIL));   /* Dump no. of bytes down chain */

/*  DPRINTF(IDL_GROSS_EVENT, ("Entered in6_ahoutput\n"));

  DP(GROSS_EVENT, outgoing->m_pkthdr.len, d);
*/

switch (flags)
{
 case SO_SENDAUTH:
                  assoc = &fixedauth[2];
                  break;
 case SO_SENDCKSUM:
                  assoc = &fixedauth[1];
                  break;

 case SO_SENDMD5:
                  assoc = &fixedauth[0];
                  break;
 case SO_SENDMD5OPT:
                  assoc = &fixedauth[3];
                  break;
 case SO_SENDAHA:
                  assoc = &fixedauth[4];
                  break;
 case SO_SENDNBOMD5OPT:
                  assoc = &fixedauth[5];
                  break;
 case SO_SENDNULLCKSUM:
                  assoc = &fixedauth[6];
                  break;
 case SO_SENDROG:
                  assoc = &fixedauth[7];
                  break;
 case SO_SENDNBOROG:
                  assoc = &fixedauth[8];
                  break;


default:  
         DAPRINTF(("Absurd flags down here\n"));
         goto ahoret;  /* CHECK */         
}

  DAP (assoc->algorithm,d);
  if (assoc->algorithm >= NAUTH)
    goto ahoret;

  algorithm = &(auth_algorithms[assoc->algorithm]);

  /* If the algorithm does not have an init function, it is an error */

  if (!algorithm->init)
    goto ahoret;

  /* Is M_WAIT okay out here ?? CHECK */
  if (!(ahm0 = ahm = m_get(M_DONTWAIT, MT_DATA)))
  {
     DAPRINTF (("ahoutput: Out of mbufs....\n"));
    goto ahoret;
  }
  ah = mtod(ahm, struct auth *);
  ah->auth_spi = assoc->spi;
  DAP (ah->auth_spi,x);
  /* check, do the different types cause a collision?? */
  ah->auth_datalen = (algorithm->datalen + 3) >> 2; /* in u_int32's */
  DAP (ah->auth_datalen,d);
  ah->auth_reserved = 0;
  /* changes , our mbufs are different */
  /* changes  , I am NOT zeroing just algorithm->datalen bytes */ 
  bzero((mtod(ahm,caddr_t) + sizeof(struct auth)), ((algorithm->datalen + 3) & (~3) ) );
  ahm->m_len = sizeof(struct auth) + ((algorithm->datalen + 3) & (~3));

  /* LOTS of changes */
  ip = mtod ( outgoing, struct ip*);  

  /* CHANGE - REMOVE later (the don't fragment flag) */
  /*ip->ip_off = ip->ip_off | IP_DF; */
  nexthdr =  ip->ip_p;
  
  /* changes CHECK -  
     Bacs - done later in the program along with length
  */

  /*      mtod(outgoing, struct ip *)->ip_off = 
	htons(mtod(outgoing, struct ip *)->ip_off);
  */

  if (nexthdr == 51)
  {
    DAPRINTF( ("ERROR, packet already had an authentication header\n"));
  }else
  {
     ah->auth_nexthdr = nexthdr;
     ip->ip_p = 51;  /* The protocol number for AH */
  }

  iphlen = mtod ( outgoing , struct ip *)->ip_hl<< 2;

  DAP (iphlen,d);
  /* CHANGE this mechanism to speed up the code? */ 
  /* CHECK - M_DONTWAIT?? */    
  MGET ( n, M_DONTWAIT, MT_HEADER);
  n0 = n;
  if (n==0)
  {   

     DAPRINTF( ("ahoutput:Failed to get an mbuf for the header\n"));
     ahoret;
  }
  
  if ( outgoing->m_len < iphlen)
  {
     DAPRINTF( ("ERROR: entire header and options were not in one mbuf"));
     ahoret;
  }
   
  outgoing->m_len -=  iphlen;

  outgoing->m_off += iphlen;
  /* ADD - release mbuf if it has become empty*/
  DAPRINTF(("After removing header, no of bytes in \"outgoing\" is\n") );
  DAP (outgoing->m_len,d);
  n->m_next = ahm;
  ahm->m_next = outgoing;
  outgoing = n;
  outgoing->m_off = MMAXOFF - iphlen; /* Putting it at the end */
  outgoing->m_len = iphlen;
  bcopy ( (caddr_t) ip, mtod (outgoing, caddr_t) , iphlen);
  
  DAPRINTF ( ("The packets have now been manipulated for firing\n"));   
  DAP ( ahm->m_len ,d);
  DAP ( outgoing->m_len,d);
  DAP ( outgoing->m_off,d);
   
    
  /*  DP(GROSS_EVENT, outgoing->m_pkthdr.len, d);
      DP(GROSS_EVENT, ahm->m_len, d);
  */

  /*  outgoing->m_pkthdr.len += ahm->m_len;
      DP(GROSS_EVENT, outgoing->m_pkthdr.len, d);
  */

  /*   IPV4_STORE_LENGTH(outgoing); Doesn't look like we need this */
  
 
  ah = mtod ( ahm, struct auth *);  /* CHECK not required? */
  ip = mtod ( outgoing, struct ip *); 
  /* increment the total datagram length to include AH and data */
  DAPRINTF (("Original length of packet"));
 
  DAP ( ip->ip_len,d);

  ip->ip_len  += ahm->m_len;
  DAP (ip->ip_len,d);
     
  /* Have to get things in NETWORK order before calculation
     ip->ip_id should  already be in network order
     ip->ip_sum has not been calculated yet but that will
     be zeroed for the AH data calculation anyways
     SWAP
  */

  ip->ip_off = htons(ip->ip_off);
  ip->ip_len = htons (ip->ip_len);  /* WHAT AN ERROR */

  DAPRINTF (("Swapped offset and length to network order\n"));  
 
  /*  DDO(IDL_GROSS_EVENT, dump_mchain(outgoing)); */

  if ( secahcalc(outgoing, assoc, algorithm, 
        ( mtod(ahm, caddr_t) + sizeof(struct auth) ) )  )
    goto ahoret;
  
/* REMOVE THIS */
#ifdef AH_DEBUG     /* to be on the safer side*/
   if (assoc->spi == htonl(0x43))
   {  
       struct auth_algorithm * unoptimized_md5;
       u_int8 buffer[32];
       struct secassoc * temp_assoc = &fixedauth[0];
  
       unoptimized_md5 = &(auth_algorithms[temp_assoc->algorithm]);

       DAPRINTF (("Checking the stuff calculated with MD5 optimized\n"));
       if ( secahcalc(outgoing ,temp_assoc, unoptimized_md5, 
            buffer ))
	 goto ahoret;
       DAPRINTF ("That was calculated by unoptimised MD5\n");
       if (bcmp(buffer, mtod (ahm,caddr_t) + sizeof(struct auth), 
	   unoptimized_md5->datalen)) 
       {
           DAPRINTF (("Both algorithms calculated DIFFERENT values!!"));
       }else
       { 
           DAPRINTF (("Both algorithms calculated the SAME values!!"));
       }
     }
#endif AH_DEBUG
 
  /* DP(GROSS_EVENT, *((u_int8 *)(ahm->m_data + sizeof(struct ipv6_auth))),
     02x);
 
  DDO(IDL_GROSS_EVENT, dump_mchain(outgoing)); */

  /* SWAP the offset and length back to host order since this will be
     done at different places again in the ip code - may CHANGE later 
  */
     
  mtod(outgoing, struct ip *)->ip_off = 
      ntohs(mtod(outgoing, struct ip *)->ip_off);
  mtod(outgoing, struct ip *)->ip_len = 
      ntohs(mtod(outgoing, struct ip *)->ip_len);



 /* CHECK Right now we already have that done, so we are not messing with it */
 
 /*

  if (mtod(outgoing, struct ip *)->ip_v == 4) 
  {
    mtod(outgoing, struct ip *)->ip_off = 
      htons(mtod(outgoing, struct ip *)->ip_off);
    mtod(outgoing, struct ip *)->ip_len = 
      htons(mtod(outgoing, struct ip *)->ip_len);
  }
 */



 /*  DP(GROSS_EVENT, outgoing->m_pkthdr.len, d); */

  DADO ( dump_chain (outgoing, DETAIL));
  DADO ( ip_header (outgoing, HOST_ORDER));
  DAPRINTF (("Exitting ahoutput() without error conditions\n"));
  return outgoing;

  ahoret:
      DAPRINTF (("ahoutput returning ERROR\n"));

     /*  DPRINTF(IDL_CRITICAL, ("in6_ahoutput returning ERROR\n")); */

 
     /*
      The following changes as this chain is probably being taken care of in
      the ip layer
     */

     /*   
       if (outgoing)
       m_freem(outgoing);  
     */

     /* 
     Since both ahm and n were prepended to the mbuf containing the header
     freeing them should still allow the ip layer to release the rest of 
     the chain
     */
          
 
      if (ahm0)
     {
        m_free (ahm0);
     }
     if (n0)
     {
        m_free (n0);
     }
     return NULL;

}
/* 
----------------------------------------------------------------------
Routine:     ahinput
Funtion:     Strip off the AH, arrange mbufs, call secahcalc 
Output:      IP datagram with AH removed
To Do :      See
            - freeing mbufs
            - see that ntoh is not done at multiple places
            - cksum wrong when returned?? 
    
----------------------------------------------------------------------
*/
/*changes : In the number of arguments to ahinput */

struct mbuf * ahinput(incoming)
     struct mbuf *incoming;
{
  struct auth *ah;
  struct secassoc *assoc;
  u_long spi;
  int counter;
  struct mbuf *datam;
  struct auth_algorithm *algorithm;
  struct ip * ip;
  int iphlen;    /* The length  of the header */ 
  struct mbuf * n;
  struct mbuf * n0 = (struct mbuf *) 0;
  int extra;     /* no of bytes over and above the  iphlen */
  int ahlength;  /* Total no. of bytes, structure + datalen */
  int pullup; /* No. of bytes to be pulled up */ 

  INIT;

  DLOG((LOG_DEBUG,"Entering ahinput(incoming=%08x)\n",incoming) );
  DAPRINTF (("Id, length, offset should all be in host order\n"));
  DAPRINTF (("The length of the header is NOT included in ip->iplen\n"));
  DADO (ip_header (incoming,HOST_ORDER));
  DADO ( dump_chain (incoming) );

  ip = mtod ( incoming, struct ip *);
  /* CHECK - network or host byte order right here? */
  
  iphlen = ip->ip_hl << 2;
  DAP ( iphlen,d );

  /*
  DP(GROSS_EVENT, ntohs((mtod(incoming, struct ipv6 *))->ipv6_length), d);
  DP(GROSS_EVENT, incoming->m_pkthdr.len, d);
  DP(GROSS_EVENT, assoc, 08x);
  */
 
  /* Leave only the header in the first mbuf, move the extra bytes into 
     a new mbuf 
  */
  /*data since this is for the AH */ 
  /* CHECK wait or dontwait */
  MGET ( n, M_DONTWAIT, MT_DATA);
  n0 = n;
  if ( n == 0)
  {  
      DAPRINTF(("ahinput:Failed to get mbuf n\n"));
      goto ahiret;
  }
  
  if ( incoming->m_len < iphlen )
  {
      DAPRINTF( ("ahinput: ERROR entire header not in one mbuf\n"));
      goto ahiret;
  } 

  DAPRINTF (("The no. of bytes in the first mbuf when it was handed in ...\n") );
  DAP (incoming->m_len,d);
  
  extra = incoming->m_len - iphlen;
  
  DAPRINTF ( ("Therefore, no. of extra bytes is ...\n"));
  
  DAP (extra,d);
  

  if (extra)
  {

    incoming->m_len -= extra; /* effectively remove extra bytes from the end */
    n->m_len = extra;  
    bcopy ( (caddr_t) (mtod (incoming, caddr_t) + iphlen), mtod ( n, caddr_t), extra) ;
  } else

  {
     n->m_len = 0;
  }
  DAPRINTF (("ahinput: After moving extra bytes ..." ));
  DAP ( incoming->m_len,d);
  DAP ( n->m_len,d);
  
  /* Now arrange the mbufs in a chain */
  
  n->m_next = incoming->m_next;
  
  /* have yet to append the chain starting at n to incoming */
  
  /* isolate incoming as of now so that if there is some error in 
     m_pullup (such as insufficient no. of bytes are there in the 
     chain and the chain is freed) then a null can be returned to
     ipintr() function (from where this has been called) and 
    it will subsequently free only the header i.e. incoming */
  /* On testing seems that m_pullup() won't function properly if the no. 
    of bytes in the current mbuf is already greater than  the no. of bytes
   to be pulled up 
  */
    
   incoming->m_next = (struct mbuf *) NULL;
   
   
    if ( n->m_len >= 2)
    {
       
       ah = mtod ( n, struct auth *);
       pullup = sizeof (struct auth) + ah->auth_datalen;
       DAP (pullup,d);
       if ( n->m_len >= pullup )
       {
         DAPRINTF (("NO pullup was required\n"));
       }else
       {
      
         if ( (n = m_pullup (n,pullup ) )  == 0)
         {
            DAPRINTF ( ("ahinput:mpullup:Packet has insufficient no. of bytes?\n"));
            goto ahiret;
         }
       
         DAPRINTF(("Only one pullup was required\n"));
       } 
    }else  /* n->m_len is less than 2 */
   {   

       DAPRINTF ( ( "Pulling up AH only (no data) first\n") );
       /* m_pullup should not fail since n->m_len is less than the size here*/
       if ( (n = m_pullup (n, sizeof (struct auth)) ) == 0)
       {
          DAPRINTF(("ahinput:mpullup:Even the ah header is not present?\n"));
          goto ahiret;
       } 
         
       ah = mtod ( n, struct auth *);
       pullup = ah->auth_datalen + sizeof (struct auth);
       
       if ( n->m_len < pullup )
       {
          DAPRINTF (("Had to be pulled up twice\n"));
          if ( (n = m_pullup ( n, pullup)) == 0 )
          {
             DAPRINTF (("ahinput: Error on the second pullup\n"));
             goto ahiret;
          }
           
       }  

    }       
   /* Now, because there were no errors, we append the modified chain
          to the initial header, incoming 
   */
          
   incoming->m_next = n;
        
   DAPRINTF (("Mbufs set up NOW \n"));   
   DADO ( dump_chain (incoming,DETAIL) );
      
   ip = mtod ( incoming, struct ip *); /* Not required ? */
   ah = mtod (n , struct auth *) ; /* What an error !!!!*/

  /* look up assoc. here */
   spi = ntohl (ah->auth_spi);  /* Should be in nw order */

   DAPRINTF (("Locating the SPI from the incoming packet\n"));
   DAP (spi,x);
   for (counter = 0; counter < NAUTH; counter++)
   {
     if ( spi == ntohl ( fixedauth[counter].spi ) )
     { 
        assoc = &fixedauth[counter];
        DAPRINTF ( ("SPI matched with that of auth algorithm no. \n"));
        DAP (counter,d);
     }
   }  
  if (!assoc)
  {
    DAPRINTF (("No match for SPI\n"));
    goto ahiret;
  }
  if (assoc->algorithm >= NAUTH)
    goto ahiret;

  algorithm = &(auth_algorithms[assoc->algorithm]);

  if (!algorithm->init)
    goto ahiret;
   
  DAP ( ah->auth_datalen,d);
  DAP ( algorithm->datalen,d);  

  /*
    DP(GROSS_EVENT, ah->auth_datalen, d);
    DP(GROSS_EVENT, algorithm->datalen, d);
  */
  if (ah->auth_datalen != ((algorithm->datalen + 3) >> 2)) 
  {
    DAPRINTF (("ahinput:The datalengths don't match\n"));
    goto ahiret;
  }
  ahlength  = (ah->auth_datalen << 2) + sizeof (struct auth);
  /* changes, CHECK - WAIT? */

  if (!(datam = m_get(M_DONTWAIT, MT_DATA)))  
  {
     DAPRINTF(("ahinput:Couldn't allocate datam mbuf\n"));
    goto ahiret;
  } 

  /*  DP(GROSS_EVENT, algorithm->datalen, d); */

  /* CHECK: DO WE NEED  to allocate only algorithm->dalalen bytes? or 
     ah->auth_datalen bytes? 
  */
  datam->m_len = algorithm->datalen;

  bzero( mtod(datam, caddr_t),( (algorithm->datalen + 3) & ~3) ); /* bacs check */
  
  DAP ( datam->m_len,d);
  DAPRINTF (("Contents of datam right now\n"));
  DADO( hex_print ( mtod (datam, caddr_t), (algorithm->datalen + 3) & ~3 ) ) ;
 
  
  /*
  DP(GROSS_EVENT, datam->m_len, d);
  DDO(IDL_GROSS_EVENT, dump_mbuf(datam));
  DDO(IDL_GROSS_EVENT, dump_mchain(incoming));
  */


  /* BACS WHAT IS this?? 
   
  */

  /*  if (mtod(incoming, struct ip *)->ip_v == 4) {
  	IPV4_STORE_LENGTH(incoming);
	mtod(incoming, struct ip *)->ip_id =
		htons(mtod(incoming, struct ip *)->ip_id);
      }

  */
   
  /* SWAP the fields in host order back into network order */
  
  /* CHECK - DO WE NEED TO THIS  AT ALL -maybe use ifdefs as in ipintr() */

   ip->ip_len = htons ( ip->ip_len);
   ip->ip_id = htons ( ip->ip_id); /* What an error !! */
   ip->ip_off = htons ( ip->ip_off); 


  if (secahcalc(incoming, assoc, algorithm, mtod(datam, caddr_t) ) )
  {
     DAPRINTF (("The security calculation failed\n"));
     goto ahiret;
  }
  

   /* SWAP the fields back into host order */
  /* ip should still be valid since we last used mtod()*/  
  ip->ip_len = ntohs ( ip->ip_len);
  ip->ip_id = ntohs ( ip->ip_id);
  ip->ip_off = ntohs (ip->ip_off);

 

 /*
  DDO(IDL_GROSS_EVENT, dump_mbuf(datam));
  DP(GROSS_EVENT, datam->m_len, d); 
  WHY was this required?
  */
  
    

  /* CHANGE when we use a finite datalength, also with the length of 
      the algorithm data

  DP(GROSS_EVENT, *((u_int8 *)((caddr_t)ah + sizeof(struct ipv6_auth))),
    02x);
  DP(GROSS_EVENT, *(mtod(datam, u_int8 *)), 02x);
  */

  /* Note, only algorithm->datalen is being compared and not ah->auth_datalen*/
  if (bcmp(mtod(datam, caddr_t), (caddr_t)ah + sizeof(struct auth), 
	   algorithm->datalen)) 
  {
 
     DAPRINTF (("AUTHENTICATION FAILED\n"));

    /* DPRINTF(IDL_CRITICAL, ("authentication failed!\n")); */

    /* If you *know* it's bad, then drop it! */
    goto ahiret;

  }

  DAPRINTF (("AUTHENTICATION  PASSED\n"));

  /*  DPRINTF(IDL_MAJOR_EVENT, ("authentication passed!\n")); */
  
  /* Were just using datam temporarily to hold data */
 
  m_free(datam);
  
  /*
  CHANGE: Whenever we modify the mbuf structure for sunos413
 
  incoming->m_flags |= M_AUTHENTIC;
  */
  /* 
  DPRINTF(IDL_GROSS_EVENT, ("Calling next\n"));
  DP(GROSS_EVENT, datam->m_len, d);
  DP(GROSS_EVENT, extra, d);
  */


  ip = mtod(incoming, struct ip *);
  
  /* for now, we must eat the AH */
  

  /*  incoming->m_pkthdr.len -= sizeof(struct ipv6_auth) + 
      (ah->auth_datalen << 2);
  */
  /* Why is this being done????? 
     Bacs - This was done in the actual IP layer for us.
     CHECK whether the secahcalc requires some modification because of
     this 
  */

  /*  ip->ip_len = incoming->m_pkthdr.len - (ip->ip_hl << 2);
      ip->ip_id = ntohs(ip->ip_id);

  */
  
  /* independent of byte ordering */	
  
  /* CHANGE - see what format the  ip_len field was in - done */
  /*  
   Modify ip_len so that it does not include the length of the 
   authentication header and the authentication data 
  */
  ip->ip_len -= ahlength;  
  ip->ip_p = ah->auth_nexthdr;

  if ( incoming->m_next != n) 
  {
     DAPRINTF (("What is going on???\n"));
     goto ahiret;
  }
  if ( (n->m_len - ahlength) != 0 )
  {
      DAP ( n->m_len,d);
      DAP (ahlength,d);
      n->m_off += ahlength;
      n->m_len -= ahlength;
      DAPRINTF (("Removed the AH bytes from the mbuf n\n"));
      DAP (n->m_off,d);
      DAP (n->m_len,d);
      
  }else 
  { 
      /* Remove the mbuf containing the AH and link the rest */
   
      incoming ->m_next = m_free (n);
      DAPRINTF (( "Removed the mbuf containing the AH altogether\n"));  
  }
  
  /* Changes, see later */  
 
  /* DDO(IDL_MAJOR_EVENT, dump_ihitab(index)); */

  /* Check, forward this here or send it back to ip?? */
  /*  (*inetsw[ip_protox[ip->ip_p]].pr_input)(incoming, extra, index, indexoff);*/
  


  DADO ( dump_chain(incoming, DETAIL));
  DADO ( ip_header (incoming, HOST_ORDER));
  DAPRINTF (("Leaving ahinput(), rval = incoming\n"));
  return incoming;

ahiret:
  DAPRINTF(("ahinput() failure\n"));
  /* CHANGE if not calling IP at all to m_freem(incoming) */
  if (n)
    m_freem(n);

  DAPRINTF(("Leaving ahinput(), rval = 0\n"));
  return NULL;
}
 













