#************************************************************************
# script.pl
# 
# Copyright (c) 1997 University of Southern California.
# All rights reserved.                                            
#                                                                
# Redistribution and use in source and binary forms are permitted
# provided that the above copyright notice and this paragraph are
# duplicated in all such forms and that any documentation, advertising
# materials, and other materials related to such distribution and use
# acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from
# this software without specific prior written permission.3
#
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

************************************************************************

#!/local/bin/perl

#-----------------------------------------------------------------
#                            TUNING PARAMETERS
$TIMES = 10;       
$GOAL  = 60;       
$blast = "../blast/blast";  
$KVAR = "/kvar";
$KSET = "/kset";
$KZERO = "/kzero";
$MAX_TCP_SIZE = 8192;     
$MAX_UDP_SIZE = 8192;
$MIN_BUF_SIZE = 128;
$INITIAL_COUNT = 2048;
$MAX_COUNT_FACTOR = 5;   #Divide the no. of packets as size increases

$MIN_PER_TICK = 10;
$MAX_PER_TICK = 80;
$TICK_INCREMENT = 1;
$COUNT_REDUCTION = 5;   #reduce count for determining optimum ticks..
@netstat_key = ("bad", "error");

$MAX_BUF_SIZE = ($MAX_TCP_SIZE <  $MAX_UDP_SIZE)? $MAX_UDP_SIZE: $MAX_TCP_SIZE;


@DEFAULT_BUF_SIZES = (8, 128, 256, 512, 1024, 2048, 4096, 8192);

#-----------------------------------------------------------------
#Parameters for plot
%extension = ("mif", ".mif", "postscript", ".ps");
$UDP_YRANGE_L = 0;
$UDP_YRANGE_H = 200;
$TCP_YRANGE_L = 0;
$TCP_YRANGE_H = 100;
#-----------------------------------------------------------------
#                            CONSTANTS
$log_file = "log.txt";
$udp_opt_log = "udp_opt_log.txt";
$udp_file = "udp";
$udp_opt_file = "udp_opt";
$tcp_file = "tcp";
$data_ext = ".data";

$udp_data_file = $udp_file.$data_ext;
$tcp_data_file = $tcp_file.$data_ext;
$udp_opt_data_file = $udp_opt_file.$data_ext;

$DIR_NAME = "obv_set";
$UDP = 1;
$TCP = 0;

%TEST_NUMBER = ("no_ah", 0,
		"null_ah", 1,
		"ck_sum", 2,
		"md5", 3,
		"md5_opt", 4,
                "aha", 5,
		"nbo_md5_opt", 6,
		"null_cksum",7,
		"rog", 8,
		"nbo_rog",9);
$ON = 1;
$OFF= 0 ; #For UDP checksum

##---------------------------------------------------------
#New data structures.

%VAR= ("t", *opt_t, "u", *opt_u, "o", *opt_o);
%SUBTESTS = ("t", "nr", "u", "sbz", "o", "sbz");  #subtests for each

#for composing file names
%PREFIX1 = ( "t", "TCP", "u", "UDP", "o", "UDP_OPT");

#for now, unique among all tests
%PREFIX2 = ( "n", "_NO_DELAY", "r", "_REGULAR", "s", "_SEND_CK", "b",
	    "_BOTH_CK", "z", "_ZERO_CK");

#-----------------------------------------------------------------
#                            GLOBALS
# LOG - descriptor for log.txt
# UDP - descriptor for UDP unoptimized data file
# TCP - descriptor for TCP data file
# UDP_OPT - descriptor for UDP optimized data file
# UDP_OPT_LOG - descriptor for log of UDP optimizations
# %do  - the tests to be done..
# $remote - the name of the remote machine..


require 'ctime.pl';
require 'getopts.pl';


#-----------------------------------------------------------------
#                            OPTIONS
# -c  : comments to put in the log
# -u  : do unoptimized UDP tests - new - specify which ones 
#       ie. checksum zeroed, at sender or both . eg. (-u zsb)
# -t  : do TCP tests (can specify regular and no_delay e.g. (-t nr)
# -b  : arguments for blast (other than those included)
# -o  : Optimize UDP receive bandwidth by using  writes per tick (contd)
#       Specify which ones (e.g. -o sbz)
# -a  : The authentication algorithms for whom tests are to be done
# -p  : Plot only, nothing else....
# -i  : Treat the hostname argument as an ip address ...
# -s  : Colon separated list of packet sizes. e.g -s 128:256:567
# 
#-----------------------------------------------------------------

&Getopts ("ipb:c:a:s:t:u:o:");
$remote = shift(@ARGV); 
&check_args ();      #check arguments
#make a dir for the script and store it in the given variable
&mk_script_dir ("$DIR_NAME.", *script_dir);  


foreach $algorithm (keys (%do)){   #once of each algorithm
    chdir ($script_dir) || die ("Can't chdir to \"$script_dir\"");
    #make directory, cd to it and open output files
    &mkdir_open_files($algorithm);  
    &log_start_info($algorithm);    
    if ($algorithm ne "no_ah"){
	if ($algorithm ne "null_ck"){
	    $blast_args = $opt_b." -h ".$TEST_NUMBER{$algorithm};
	    $sp_blast_args = " ";
	}
	else {
	    $blast_args = $opt_b." -h 2" ;
	    $sp_blast_args = $opt_b." -h 7";
	}
    }

    &start_remote_blast();     #Need only do once??
    
#    for ($length = $MIN_BUF_SIZE, $count = $INITIAL_COUNT;
#	 $length  <= $MAX_BUF_SIZE; 
#	 $length *= 2, $count = sprintf ("%.0f", $count/2)){
#	
#	$count = INITIAL_COUNT;
#	foreach $length (@packet_size_list){
#	    $count 

    for ($index = 0, $count = $INITIAL_COUNT;
        $index <= $#packet_size_list;
	 $index++,
	 $count = int ($count/$factor)){

	$length = $packet_size_list[$index];
	$factor = $packet_size_ratios {$length};
	unless($factor){  #for the last time around
	    $factor = 1;   
	}

	if ($opt_t && ($length <= $MAX_TCP_SIZE)) {  #Do a TCP iteration..
	    
	    foreach $subtest (@opt_t){  #presently r and n
		if ($subtest eq "r"){
		    $tcp_blast_args = $blast_args; #regular tcp
		}elsif ($subtest eq "n"){ #no delay
		    $tcp_blast_args = $blast_args. " -n ";
		}
		$count = &update_count($TCP, $count, $length, $tcp_blast_args);
		&get_tcp_data($count, $length, $tcp_blast_args, 
			      $PREFIX2{$subtest}, *tcp_data);
		$filename = &data_file_name ("t", $subtest);
		print $filename ($tcp_data);
	    }
	}
	if ( ($opt_u || $opt_o) && ($length <= $MAX_UDP_SIZE)){

#count was here before - check

	    if ($opt_u){
		foreach $subtest (@opt_u){
		    &configure_udp_cksums ($subtest);
		    $count = &update_count ($UDP, $count, $length, $blast_args);
		    &get_udp_data ($count, $length, $blast_args, 
				   $PREFIX2{$subtest}, *udp_data, 0);

		    $filename = &data_file_name ("u", $subtest);
		    print $filename ($udp_data);
		}
	    }
	    if ($opt_o){
		foreach $subtest (@opt_o) {
		    &configure_udp_cksums ($subtest);
		    $count = &update_count ($UDP, $count, $length, $blast_args); 
		    $writes_per_tick = &get_optimum_wpt ($count, $length, $blast_args, $subtest);
#		$count = &update_count ($UDP, $count, $length, $blast_args." -w $writes_per_tick");
		    $filename = &data_file_name ("o", $subtest);
		    print $filename ("#$writes_per_tick\n");

		    &get_udp_data ($count, $length, $blast_args, 
			$PREFIX2{$subtest}, *udp_opt_data, $writes_per_tick);

		    print $filename ($udp_opt_data);
		}
	    }

	}   
    }
    &close_files();
}

#go ahead and plot these files...

&plot ($script_dir);


#-----------------------------------------------------------------
#Name :    get_optimum_wpt()
#Function: Find the optimum number of ticks for the packetsize and blast_args
#Inputs :  Number of buffers ($count), length of each buffer ($length)
#          Any other blast options ($opt_b), $subtest has checksum options
#Ouputs :  Returns the optimum number of tics..
#Comments: Logs to a filename composed from the subtests
#-----------------------------------------------------------------
sub get_optimum_wpt{
    local ($count, $length, $blast_args, $subtest) = @_;
    local ($opt_wpt, $max_recv, $wpt, $old_fh, @blast_out, @blast_line, $command, $filename);

    $count = sprintf ("%.0f", $count/$COUNT_REDUCTION);
    &print_time (" Finding optimum writes per tick for $PREFIX2{$subtest} sz = $length, args = \"$blast_args\", see the other log\n"); 
    $filename = &get_file_name ("o", $subtest, "log");
    $old_fh = select ($filename); $| = 1;
    &print_time (sprintf ("UDP_OPT $PREFIX2{$subtest} : sz = %d, cnt (actual) = %d, args = \"%s\"", $length, $count, $blast_args));    

    for ($wpt = $MIN_PER_TICK; $wpt < $MAX_PER_TICK; $wpt += $TICK_INCREMENT){

	$command = "$blast -u -l $length -c $count -w $wpt $blast_args $remote";
	print ("$command \n");
	open (PERF, "$command |")||
		  die ("$0: get_optimum_wpt(): Can't open blast...");
	@blast_out = ();
	@blast_out = <PERF>;
	close (PERF);
	print @blast_out;
	print ("\n");

LINE:
	foreach $_ (@blast_out) {
	    if (/^[\s]*Time/){
		@blast_line = split (" ", $_);
		if ($max_recv < $blast_line[5]){
		    $max_recv = $blast_line[5];
		    $opt_wpt = $wpt;
		}
		# Now adjust the count!!
		$count *= ($GOAL/$blast_line[1]);
		$count = sprintf ("%.0f", $count);
		last LINE;
	    }
	}
	if (@blast_line <= 0) {
	    print STDERR ("$0:get_optimum_wpt() : No reading.. \n");
	    exit (1);
	}

    }

    &print_time("Optimum writes per tick for sz = $length is $opt_wpt\n");
    print ("************************************************************");
    select ($old_fh);
    print ("Optimum writes per tick for sz = $length is $opt_wpt\n");
    return ($opt_wpt);
}
#-----------------------------------------------------------------
#Name :    get_udp_data()
#Function: Do $TIMES iterations with $blast for tcp
#Inputs :  Number of buffers ($count), length of each buffer ($length)
#          Any other blast options ($opt_b), reference to data. (*udp_data)
#          $wpt - which is the number of writes per tick if any... 
#Ouputs :  Puts the results of the iteration in *udp_data
#Comments: Logs to currently selected filehandle.
#          TODO : Call rec_netstat to get changes in statistics.
#-----------------------------------------------------------------
sub get_udp_data {
    local ($count, $length, $blast_args, $subtest_ident, *data, $wpt) = @_;
    local (@blast_out, @blast_line, @send_line, @recv_line,$comment);
    
    if ($wpt != 0) {
	$blast_args.=" -w $wpt";
	$comment = "OPTIMIZED";
    }
    &print_time (sprintf("UDP $comment $subtest_ident : sz = %d, cnt = %d", $length, $count));
    &rec_netstat ("(rsh $remote netstat -s)|", *before);
    open (PERF, "$blast -u  -i $TIMES -c $count -l $length $blast_args $remote|") || die ("$0:get_udp_data - Can't start blast....");
    @blast_out = <PERF>;
    close (PERF);
    print @blast_out;
    foreach $_ (@blast_out) {
	if (/^RCV\s/) {
	    @recv_line = split (" ", $_);
	}elsif (/^SND\s/){
	    @send_line = split (" ", $_);
	}
    }
    if ( (@send_line <= 0) || (@recv_line <= 0) ) 
    {
	print STDERR ("$0: get_udp_data() - No blast reading, exitting..\n");
	exit (1);
    }
    chop ($recv_line[5]);      #RCV Best
    chop ($recv_line[7]);      #RCV Worst
    chop ($recv_line[9]);      #RCV Average
    chop ($send_line[5]);      #SND Best
    chop ($send_line[7]);      #SND Worst
    chop ($send_line[9]);      #SND Average
    
    $loss = $send_line[9] - $recv_line[9];

    #    format : length (send, recv) (average sd max min)  loss

    $data = sprintf ("%d %0.2f %0.2f %0.2f %0.2f %0.2f %0.2f %0.2f %0.2f %0.2f \n",
	     $length, 
	     $send_line[9], $send_line[11], $send_line[5], $send_line[7], 
	     $recv_line[9], $recv_line[11], $recv_line[5], $recv_line[7],
	     $loss);
    &rec_netstat ("(rsh $remote netstat -s)|", *after);
    &print_b_and_a ("UDP $comment blast recv", *before, *after);
}

#-----------------------------------------------------------------
#Name :    get_tcp_data ()
#Function: Do $TIMES iterations with $blast for tcp
#Inputs :  Number of buffers ($count), length of each buffer ($length)
#          Any other blast options ($opt_b), and reference to data. (*tcp_data)
#Ouputs :  Puts the results of the iteration in *tcp_data
#Comments: Logs to currently selected filehandle.
#          TODO : Call rec_netstat to get changes in statistics.
#-----------------------------------------------------------------
sub get_tcp_data {
    local ($count, $length, $blast_args, $subtest_ident,*data) = @_;
    local (@blast_out, @blast_line);
    &print_time (sprintf ("TCP $subtest_ident: sz = %d, cnt = %d", $length, $count));
    &rec_netstat ("(rsh $remote netstat -s)|", *before);
    open (PERF, "$blast -i $TIMES -c $count -l $length $blast_args $remote|") || die ("$0:get_tcp_data - Can't open blast.");
    @blast_out = <PERF>;
    close (PERF);     #redundant..
    print (@blast_out);
    foreach $_ (@blast_out){
	if (/^RCV\s/){
	    @blast_line = split (" ", $_);
	    
	    chop ($blast_line[5]);     # Best
	    chop ($blast_line[7]);     # Worst
	    chop ($blast_line[9]);     # Average
	    #format: length average standard_dev max min
	    $data = sprintf ("%d %0.2f %0.2f %0.2f %0.2f\n", $length, $blast_line[9], $blast_line[11], $blast_line[5], $blast_line[7]);
	    last;
	}
    }
    if (@blast_line <= 0) {
	print STDERR ("$0: get_tcp_data() - No blast reading, exitting..\n");
	exit (1);
    }
    &rec_netstat ("(rsh $remote netstat -s)|", *after);
    &print_b_and_a ("TCP blast recv", *before, *after);
}

#-----------------------------------------------------------------
#Name :    start_remote_blast()
#Function: Start whatever blast required at remote machine
#Inputs :  None
#Ouputs :  None
#Comments: TODO : Update to check that blast actually started without error..
#          Log to currently selected filehandle..
#          Directs STDERR to /dev/null and restores it to avoid unneeded 
#          comments..
#-----------------------------------------------------------------
sub start_remote_blast {
    &print_time (" Starting remote blast ....\n");
    open (SAVEERR, ">&STDERR");
    open (STDERR, "/dev/null");
    if ($opt_t){
	system "rcmd $remote "."$blast $sp_blast_args";
    }
    if ($opt_u || $opt_o){
	system "rcmd $remote "."$blast -u";
    }
    open (STDERR, ">&SAVEERR");
    close (SAVEERR);
}

#-----------------------------------------------------------------
#Name :    update_count();
#Function: Update the count to last GOAL seconds
#          for the log and select it..
#Inputs :  Current count, whether TCP or UDP, and current packet length
#Ouputs :  the updated count...
#Comments: Exits if no reading found....., 
#          Prints to STDOUT, have to select before if desire otherwise.
#-----------------------------------------------------------------

sub update_count{
    local ($udp_tcp, $count, $length, $blast_args) = @_;
    local ($command, @blast_out, $secs);

    if ($udp_tcp == $UDP) {
	$blast_args .= " -u";
    }
    $command = "$blast -c $count -l $length $blast_args $remote";
    print ("******* Beginning Update count *******\n");
    print ("Buffer length: $length, Current count = $count\n");
    print ("$command \n");
    open (COMMAND, "$command|") || die (" Can't open local blast...\n");
    @blast_out = <COMMAND> ;
    close (COMMAND);
    print @blast_out;
    $secs = undef;
    foreach $_ (@blast_out)
    {
	if (/^\sTime:/){
	    $secs = (split (" "))[1];
	    last;
	}
    }
    if (!defined $secs) {
	print STDERR ("$0: update_count(): No blast reading - exitting..\n");
	exit (1);
    }
    
    $count *= ($GOAL/$secs);
    $count = sprintf ("%.0f", $count);

    print ("Time taken = $secs seconds\n");
    print ("Revised count = $count\n");
    print ("******* Update over *******\n");
    return $count;
}

#-----------------------------------------------------------------
#Name :    log_start_info();
#Function: Log the comments, kernel info on both sides and 
#          start time.           
#          for the log and select it..
#Inputs :  None..
#Ouputs :  
#Comments: Assumes that proper filehandle selected...
#-----------------------------------------------------------------
sub log_start_info {
    local ($algo) = @_;
    print ("$algo\n");
    print "Start: ", &ctime(time), $opt_c, "\n";
    print "$internal_c";
    #Add some kernel comments
    chop ($_ = `hostname`);
    print ("Source: ", $_,"\n",`cat /etc/motd` , "Destination: ", $remote, "\n", `rsh $remote cat /etc/motd`), "\n\n";
}
    

#-----------------------------------------------------------------
#Name :    print_time($str)
#Function: Print the time with the string passed in..
#Inputs :  String to printed with the time..
#Ouputs :  None
#-----------------------------------------------------------------
sub print_time{
    local($str) = @_;	
    $tmp = &ctime(time);
    chop($tmp);
    print "\nTime: ", $tmp, "; $str\n";
}
#-----------------------------------------------------------------
#Name :    mk_script_dir() 
#Function: Make a directory for the script (with current time).
#Inputs :  The first part of the name, and a pointer wherein to 
#          place the full path...
#Ouputs :  
#Comments: Cd's to that directory on return..
#-----------------------------------------------------------------
sub mk_script_dir {
    local ($dir, *path) = @_;
    $dir.= time;
    mkdir ($dir, 0755) || die ("$0: Can't make $dir ");
    chdir ($dir) || die ("$0: Can't cd to $dir");
    chop($path = `pwd`);

}

#-----------------------------------------------------------------
#Name :    close_files()
#Function: Closes all open file_handles...
#Inputs :  None
#Ouputs :  None
#-----------------------------------------------------------------
sub close_files {
    close (LOG);
    if ($opt_u){
	close (UDP);
    }
    if ($opt_o) {
	close (UDP_OPT);
	close (UDP_OPT_LOG);
    }
    if ($opt_t) {
	close (TCP);
    }
}
#-----------------------------------------------------------------
#Name :    mkdir_open_files ()
#Function: Make a directory of a particular name .
#          cd to it and open the file descriptors
#          required, seeing the options. Make output unbuffered
#          for all. Selects the log.....
#Inputs :  The name of the directory..
#Ouputs :  None..
#Comments: Is in the new directory once done...
#-----------------------------------------------------------------
sub mkdir_open_files {
    local ($dir) = @_ ;
    local (*variable, $test, $filename);       #Assign $opt_u, $opt_t and $opt_o in turn..

    mkdir ($dir, 0755) || die "Can't make $dir..";
    chdir ($dir) || die "Can't cd to dir $dir..";
    
    open (LOG, ">$log_file") || die "Can't open $log_file...";
    
    while ( ($test, *variable) = each %VAR){  #u, t and o currently
	if ($variable) { 
	    #@variable has also been initialized then
	    foreach $_ (@variable) {  #i.e. each subtest to do..
		$filename = &data_file_name ($test, $_);
		open ($filename, ">$filename") || die ("Can't open $filename $!\n");
		select ($filename) || die ("Can't select!\n");
		$| = 1;
		
		#open the logs for optimization
		if ($test eq "o") {
		    $filename = &get_file_name ($test, $_, "log");

		    open ($filename, ">$filename") || die ("Can't open $filename $!\n");
		    select ($filename) || die ("Can't select!\n");
		    $| = 1;
		}

	    }
	}
    }
    # make unbufferred
    select (LOG) ; $| = 1;
}

#-----------------------------------------------------------------
#Name :    check_args()
#Function: Check arguments passed in to script, set defaults if reqd.
#          Check that remote host exists..
#Inputs :  All global
#Ouputs :  None, exits if not valid..
#-----------------------------------------------------------------
sub check_args {

    local ($algos, $test, *variable);
    if (!$opt_p) {
	if (!$remote){    #Need remote machine 
	    &usage () ;
	}else {            #Check valid host 
	    unless ($opt_i) { #is a name
		if ( (gethostbyname ($remote))[0] eq "") {
		    print STDERR ("$0: remote host \"$remote\" doesn't exist??\n");
		    exit(1);
		}
	    }else
	    {
		if ( (gethostbyaddr( pack ('C4', split (/\./, $remote)),2))[0] eq ""){
		    print STDERR ("$0: remote ip address \"$remote\" doesn't exist??\n");
		    exit (1);
		}
	    }
	}
    }
    #If neither tcp or udp optimized nor udp unoptimized specified
    if (!($opt_u || $opt_t || $opt_o)){  #do all with all subtests
	$opt_o = $SUBTESTS{"o"};
	$opt_u = $SUBTESTS{"u"};
	$opt_t = $SUBTESTS{"t"};
    }
    # The default test to be done..

    if ($opt_a) {    #the authentication algorithms to be done..
	@algos =  split (":" , $opt_a);
	if (@algos <= 0) {
	    print STDERR ("$0: check_args(): Invalid algorithms...\n");
	    exit (1);
	}
	foreach $_ (@algos) {
	    if (defined ($TEST_NUMBER{$_})){
		$do{$_} = 1;
	    }else {
		print STDERR  ("$0: check_args(): Invalid algorithm \"$_\"\n");
		exit(1);
	    }
	}
    }else{
	%do = ("nbo_md5_opt", 1);    
    }
    
    #Set internal comments 
    $internal_c.="Measuring: \n";

    while (($test, *variable) = each %VAR){
    	if ($variable){ #make sure valid options and none repeated
	    if ( !($variable =~ /^[$SUBTESTS{$test}]+$/)  ||
		($variable =~ /([$SUBTESTS{$test}])[^1]*\1/)){
#		print ("1 is $1 2 is $2 3 is $3\n");
		print STDERR ("$0: check_args() :Incorrect option - $variable \n");
		&usage();
	    }
	    $internal_c.= "\t$PREFIX1{$test}";
	    @variable = split ("", $variable);
	    foreach $_ (@variable){
		$internal_c.= "\t$PREFIX2{$_}";
	    }
	    $internal_c.="\n";
	}
    }
#    print ("$internal_c\n"); #DEBUG
#    exit(0);

    if ($opt_p){  # Plot only....
	&plot("./"); #Files should be in the current directory...
	exit (0);
    }
    if ($opt_s) {  #Packet size list
	&expand_check_list ($opt_s, *packet_size_list);

    }else {
	@packet_size_list = @DEFAULT_BUF_SIZES;
    }
    &ratio_list (*packet_size_list, *packet_size_ratios);

#    DEBUG    
#    @temp = %packet_size_ratios;
#    print (@temp);
#    exit(0);

}


#-----------------------------------------------------------------
#Name :    expand_check_list 
#Function: Expand, check and sort packet size list.
#Inputs :  variable ($opt_s) containing colon separted values
#          destination list.
#Ouputs :  
#Comments: Will accept stuff like 10hello as equal to 10
#Does not check for duplicate packet_sizes in the list itself.
#-----------------------------------------------------------------
sub expand_check_list  {
    local ($input, *destination_list) = @_;
    local (@temp, $actual, $truncated) ;
    
    @temp = split (":", $input);
    
    foreach $actual (@temp) {
	$truncated = sprintf ( "%.0f", $actual);
	if ( ($truncated!= $actual) || ($truncated <= 0) || ($truncated >= $MAX_BUF_SIZE)) 
	{
	    print STDERR (" Improper numerical list entered\n");
	    &usage();
	}
	push (@destination_list, $truncated);
    }
    #values are above board - now sort them
    
    @destination_list   =   sort numerically (@destination_list);
}
    
sub numerically { $a <=> $b };

#-----------------------------------------------------------------
#Name :    ratio_list
#Function: Make a ratio associative array
#Inputs : Number list and the ratio list for the output
#Ouputs :  
#Comments: Check for duplicate packets done here...
#-----------------------------------------------------------------

sub ratio_list {
    local (*number_list, *ratio_list) = @_;
    local ($i, $element);
    
    for ($i = 0; $i < $#number_list; $i++) {
	$element = $number_list[$i];

	if ($ratio_list {$element}){
	    print STDERR ("ratio_list(): Duplicate numbers in list\n");
	    &usage();
	}
	$ratio_list {$element} = int ($number_list[$i+1]/$element);
	if ($ratio_list{$element} > $MAX_COUNT_FACTOR)
	{ 
	    $ratio_list{$element} = $MAX_COUNT_FACTOR;
	}
    }
    if ($ratio_list {$number_list[$#number_list]}) {
	print STDERR ("ratio_list(): Duplicate numbers in list\n");
	&usage();
    }
}


#-----------------------------------------------------------------
#Name :    usage()
#Function: Print message on usage to STDERR, exit..
#Inputs :  None
#Ouputs :  None
#-----------------------------------------------------------------

sub usage { print STDERR ("Usage: $0 [-t [n][r]] [-u [s][b][z]] [-o[s][b][z]] [-c \"comments\"] [-b \"blastargs\"] [-a no_ah[:null_ah][:md5][:md5_opt][:nbo_md5_opt][:aha][:ck_sum][:null_cksum][:rog][:nbo_rog]] [-s num1[:num2][:num3]..]  [remote_host | -p ]\n"); exit (1); }

#-----------------------------------------------------------------
#Name :    rec_netstat()
#Function: Record netstat resutls so that before and after may be compared
#Inputs :  a command for netstat and a variable to store output in
#Ouputs :  
#Comments: Save only results that contain a keyword from @netstat_key
#-----------------------------------------------------------------

sub rec_netstat {
    local($cmd, *name) = @_;
    open(TMP, $cmd);
    
    $i = 0;
    while (<TMP>) {
	if (/^\S/) {
	    chop;
	    $head = $_;
	}
	else {		
	  KEYS:	
	    foreach $netstat_key (@netstat_key) {
		if ($_ =~ $netstat_key) {
		    $name[$i++] = $head . $_;
		    last KEYS;
		}
	    }
	}
    }
    close(TMP);
}
#-----------------------------------------------------------------
#Name :    print_b_and_a()
#Function: Print any netstat statistics that have changed...
#Inputs :  A title, and before and after arrays...
#Ouputs :  
#Comments: Prints to currently selected FH..
#-----------------------------------------------------------------

sub print_b_and_a {
    local($title, *before, *after) = @_;
    $first = 1;
    for ($i = 0; $before[$i]; $i++) {
	if ($before[$i] ne $after[$i]) {
	    if ($first) {
		print $title, " netstat stats:\n";
		$first = 0;
	    }
	    local(@rest) = split(" ", $before[$i]);
	    local(@rest1) = split(" ", $after[$i]);
	    for ($j = 0; $rest[$j]; $j++) {
		if ($rest[$j] =~ /\d/) {
		    $rest[$j] .= ("(".$rest1[$j].")");
		}
	    }
	    print "\t", join(" ", @rest), "\n";
	}
    }
}

#-----------------------------------------------------------------
#Name :    get_file_name
#Function: Returns the name of the file with the given extension put in
#Inputs :  test in form of t, u, o && subtest as nr sbz
#Ouputs :  The name of the file
#Comments: The extension should not contain a dot
#-----------------------------------------------------------------
sub get_file_name {
    local ($test, $subtest, $extension) = @_;
    local ($filename);

    $filename = $PREFIX1{$test};
    $filename.= $PREFIX2{$subtest};
    $filename.=".";
    $filename.=$extension;
    return ($filename);
}

#----------------------------------------------------------------
#Name :    data_file_name
#Function: Returns the name of the data file for that test/subtest
#Inputs :  test in form of t, u, o && subtest as nr sbz
#Ouputs :  The name of the data file
#Comments: 
#-----------------------------------------------------------------
sub data_file_name {
    local ($test, $subtest)  = @_;
    local ($filename); 
    $filename = $PREFIX1{$test};
    $filename.= $PREFIX2{$subtest};
    $filename.= $data_ext;
    return ($filename);
}


#-----------------------------------------------------------------
#Name :    configure_udp_cksums
#Function: Configures local and remote udp_checksums
#Inputs :  Just the option  - s, z or n
#Ouputs :  Crash on failure...!! CHANGE!!
#Comments: uses local_udp_cksum and remote_udp_cksum...
#-----------------------------------------------------------------
sub configure_udp_cksums {
    local ($key) = @_;
    if ($key eq "z") { #no checksum at either
	&local_udp_cksum ($OFF);
	&remote_udp_cksum ($OFF);
    }elsif ($key eq "s"){ #only sender computes checksum
	&local_udp_cksum ($ON);
	&remote_udp_cksum ($OFF);
    }elsif ($key eq "b") { #both compute checksums
	&local_udp_cksum ($ON);
	&remote_udp_cksum ($ON);
    }else {
	die ("$0: configure_ck_sum() : Incorrect option passed in\n");
    }
	   
}

#-----------------------------------------------------------------
#Name :    local_udp_cksum 
#Function: Changes and confirms change of local UDP checksum calc.
#Inputs :  $ON or $OFF
#Ouputs :  
#Comments: TODO
#-----------------------------------------------------------------
sub local_udp_cksum {
    local ($onoff ) = @_;
    &udp_cksum ($onoff , ""); #"" means local host..
    $onoff = $onoff ? "ON": "OFF";
    print ("LOCAL UDP CHECKSUM $onoff NOW\n");    
}

#-----------------------------------------------------------------
#Name :    local_udp_cksum 
#Function: Changes and confirms change of local UDP checksum calc.
#Inputs :  $ON or $OFF
#Ouputs :  
#Comments: TODO
#-----------------------------------------------------------------


sub udp_cksum {
    local ($onoff, $at_host) = @_;
    local ($string, $program, $pattern, $variable, $command);


    if ($onoff) {
	$program = "$KSET";
	$program_pattern = '_udp_cksum(.*)one';
	$kvar_pattern = '_udp_cksum\sValue:\s1';
    }else {
	$program = "$KZERO";
	$program_pattern = '_udp_cksum(.*)zero';
	$kvar_pattern = '_udp_cksum\sValue:\s0';
    }

    $command = "$at_host $program _udp_cksum";
    $program_out = `$command`;
    print "Executing \"$command\"\n";
    print $program_out;
    $command = "$at_host $KVAR _udp_cksum";
    $kvar_out = `$command`;
    print "Executing \"$command\"\n";
    print $kvar_out;

    unless (&pattern_present ($program_out, $program_pattern) && &pattern_present ($kvar_out, $kvar_pattern)) {

	print STDERR ("$0: _udp_cksum could not be modified\n");
	print ("$0: _udp_cksum could not be modified\n");
	exit (1);
    }
}

##-----------------------------------------------------------------
#Name :    pattern_present
#Function: Checks that the second pattern is present in the first string
#Inputs :  a string and a  pattern
#Ouputs :  
#Comments:  use multi-line pattern instead?
#-----------------------------------------------------------------
sub pattern_present {
    local ($string, $pattern) = @_;
    local (@string);

    @string = split ("\n", $string);
    foreach $_ (@string) {
	if ( /$pattern/) {
	    return(1);
	}
    }
    return (0);
}

##-----------------------------------------------------------------
#Name :    remote_udp_cksum 
#Function: Changes and confirms change of remote UDP checksum calc.
#Inputs :  $ON or $OFF
#Ouputs :  
#Comments: TODO (Use kvar is the simple way)
#-----------------------------------------------------------------
sub remote_udp_cksum {
    local ($onoff) = @_;
    &udp_cksum ($onoff, "rsh $remote");
    $onoff = $onoff? "ON":"OFF";
    print ("REMOTE UDP CHECKSUM $onoff NOW\n");
}
#-----------------------------------------------------------------
#Name :    option_present (
#Function: Checks if the second string is present in the first one.
#Inputs :  A variable and an option
#Ouputs :  1 if present, 0 otherwise
#Comments: 
#-----------------------------------------------------------------
sub option_present {
    local ($variable, $option) = @_;
    if ($variable =~ /$option/){
	return (1);
    }else {
	return (0);
    }
}

#-----------------------------------------------------------------
#Name :    plot()
#Function: Plot all the tests taken..
#Inputs :  (Directory to go to and and run the plot...- usually $script_dir)
#Ouputs :  None
#Comments: 
#-----------------------------------------------------------------
sub plot {
    local ($dir)  = @_;
    local ($algo, $first, $output_file, $terminal, $point_style, $line_style, $count, $title, $data_file , $file) ;

    chdir ($dir) || die ("Can't chdir to \"$dir\" ");

    #Check that all the files exist...
    foreach $algo (keys (%do)){
	if ($opt_u) {
	    (-r "$algo/$udp_data_file")  || die ("Can't find $udp_data_file in $algo");
	}
	if ($opt_p) {
	    (-r "$algo/$udp_opt_data_file") || die ("Can't find $udp_opt_data_file in $algo");
	}

	if ($opt_t) {
	    (-r "$algo/$tcp_data_file" ) || die ("Can't find $tcp_data_file in $algo");
	}
    }

    open (GNUPLOT, ">gnuplot") || die ("Can't open gnuplot...");
    print GNUPLOT ("set logscale x 2\n");
    print GNUPLOT ("set xlabel \"Pkt Size (bytes)\"\n");
    print GNUPLOT ("set ylabel \"Transfer Rate (Mbits/sec)\"\n");
    
    #TODO - change this..
    print GNUPLOT ("set nokey\n");

    foreach $terminal (keys (%extension))
    { 
	
	print GNUPLOT ("set term $terminal\n");
	
	#First plot TCP for all algorithms....

	if ($opt_t) {
	    $output_file = $tcp_file.$extension{$terminal};

	    print GNUPLOT ( "set output \"$output_file\"\n");
	    print GNUPLOT ( "set title \"TCP blast data\"\n");
	    print GNUPLOT ( "set xrange [$MIN_BUF_SIZE:$MAX_TCP_SIZE]\n");   
	    print GNUPLOT ( "set yrange [$TCP_YRANGE_L:$TCP_YRANGE_H]\n");   

	    $first = 1;   #the first test for this  plot???
	    $point_style = 1;
	    $line_style = 1;
	    
	    foreach $algo (keys %do) {
		unless ($first) {
		    print GNUPLOT (", ");
		}else {
		    print GNUPLOT ("plot ");
		    $first = 0;
		}
		#max, avg, min, sd
		print GNUPLOT ("\"$algo/$tcp_data_file\" using 1:4  w lines $line_style $point_style, \"$algo/$tcp_data_file\" using 1:2 w lines $line_style $point_style, \"$algo/$tcp_data_file\" using 1:5 w lines $line_style $point_style, \"$algo/$tcp_data_file\" using 1:2:3 w errorb $line_style $point_style ");
		$line_style++;
		$point_style++;
	    }
	    print GNUPLOT ("\n");
	}

	
	# Now for UDP

UDP_PLOT:
	for ($count = 0; $count <= 1; $count++){
	    if ($count){
		unless ($opt_u) { next UDP_PLOT};
		$file = $udp_file;
		$title = "UDP (unoptimized) blast ";
	    }else {
		unless ($opt_o){ next UDP_PLOT};
		$file = $udp_opt_file;
		$title = "UDP (optimized) blast ";
	    }
	

	    
	    # for SEND first..
	    $output_file = $file."_snd".$extension{$terminal};
	
	    print GNUPLOT ( "set output \"$output_file\"\n");
	    print GNUPLOT ( "set title \"$title send data\"\n");
	    print GNUPLOT ( "set xrange [$MIN_BUF_SIZE:$MAX_TCP_SIZE]\n");   
	    print GNUPLOT ( "set yrange [$UDP_YRANGE_L:$UDP_YRANGE_H]\n");   

	    $point_style = 1;
	    $line_style = 1;
	    $first = 1;
	    
	    foreach $algo (keys (%do) )
	    {
		unless ($first) {
		    print GNUPLOT (", ");
		}else {
		    print GNUPLOT ("plot ");
		    $first = 0;
		}
		
		$data_file = $algo."/".$file.$data_ext;
		
		print GNUPLOT " \"$data_file\" using 1:4 with lines $line_style $point_style, \"$data_file\" using 1:5 w lines $line_style $point_style, \"$data_file\" using 1:2 w lines $line_style $point_style, \"$data_file\" using 1:2:3 w errorb $line_style $point_style ";
		$line_style++;
		$point_style++;


		
	    }
	    
	    print GNUPLOT ("\n");
	    # for RECV now
	    $output_file = $file."_rcv".$extension{$terminal};
	
	    print GNUPLOT ( "set output \"$output_file\"\n");
	    print GNUPLOT ( "set title \"$title recv data\"\n");
	    print GNUPLOT ( "set xrange [$MIN_BUF_SIZE:$MAX_TCP_SIZE]\n");   
	    print GNUPLOT ( "set yrange [$UDP_YRANGE_L:$UDP_YRANGE_H]\n");   

	    $point_style = 1;
	    $line_style = 1;
	    $first = 1;
	    
	    foreach $algo (keys (%do) )
	    {
		unless ($first) {
		    print GNUPLOT (", ");
		}else {
		    print GNUPLOT ("plot ");
		    $first = 0;
		}
		$data_file = $algo."/".$file.$data_ext;
		print GNUPLOT " \"$data_file\" using 1:8 with lines $line_style $point_style, \"$data_file\" using 1:9 w lines $line_style $point_style, \"$data_file\" using 1:6 w lines $line_style $point_style, \"$data_file\" using 1:6:7 w errorb $line_style $point_style ";
		$line_style++;
		$point_style++;
		
	    }
		
	    print GNUPLOT "\n";

	}  # end both the UDP stufff..

    }

}


