/*************************************************************************
 *									 *
 * ATOMIC UDP/TCP Performance Measurement Program (blast.c)		 *
 * December, 1995							 *
 *									 *
 * Copyright (c) 1995, 1996, 1997 by the University of Southern California*
 * All rights reserved.							 *
 *									 *
 * Permission to use, copy, modify, and distribute this software and its *
 * documentation in source and binary forms for non-commercial purposes	 *
 * and without fee is hereby granted, provided that the above copyright	 *
 * notice appear in all copies and that both the copyright notice and	 *
 * this permission notice appear in supporting documentation, and that	 *
 * any documentation, advertising materials, and other materials related *
 * to such distribution and use acknowledge that the software was	 *
 * developed by the University of Southern California, Information	 *
 * Sciences Institute.  The name of the University may not be used to	 *
 * endorse or promote products derived from this software without	 *
 * specific prior written permission.					 *
 *									 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about	 *
 * the suitability of this software for any purpose.  THIS SOFTWARE IS	 *
 * PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,	 *
 * INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF		 *
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.		 *
 *									 *
 * Other copyrights might apply to parts of this software and are so	 *
 * noted when applicable.						 *
 *									 *
 *************************************************************************/
/*                            blast.source_route
This program includes the alignment modifications made by Annette in 
Dec 95 and can measure TCP bandwidth with loose and strict source
routing 
Put the variables avndebug  and avndebug2 both to zero if you need the
program to do source routing silently.

*/

/* $Id
   $Log
*/

#include <stdio.h>
#include <ctype.h>
#include <netdb.h>
#include <signal.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <net/if.h>
#include <netinet/tcp.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>



/* source routing  begin*/
#ifdef _AIX
#include <sys/select.h>
#endif
#include <sys/file.h>
#include <sys/ioctl.h>

#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/ip_icmp.h>
#include <netinet/udp.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
/* source routing end*/



#ifdef ATP
#include "sock_atp.h"
#endif /* ATP */

#ifndef FALSE
#define FALSE            0   /* logical false */
#define TRUE             !FALSE	/* logical true */
#endif
#define boolean int

#define GET_TIME(t_val) gettimeofday(&(t_val), 0) 

#define KBITS_PER_SEC(bytes, usecs) ((bytes*125)/(usecs>>6))

#define SUB_TIME(t,sub) {               \
                t.tv_sec-=sub.tv_sec;   \
                t.tv_usec-=sub.tv_usec; \
		if (t.tv_usec < 0) {    \
                    t.tv_sec--;         \
                    t.tv_usec+=1000000; \
                }                       \
                if (t.tv_sec < 0) {     \
                    t.tv_sec = 0;       \
                    t.tv_usec = 0;      \
                    } \
		}

void tcp_listen(), tcp_send(), udp_listen(), udp_send();

/* source routing begin*/

/*
   Might have to align the optlist on a 4 byte boundary
   TO CHECK: Probably not required
*/


/* The following variables have been kept global for now */
/* AH */

/* Set option to send packets without the authentication header */
int authenticate = FALSE;
int nodelay_on = FALSE;  
/* stuff to make the setsockopts etc easier*/
typedef struct
{
  u_int option_value;
  char * option_name;
} option;

option options[] = 
{
{SO_SENDAUTH, "INVALID OPTION"},         /* option number 0 -not valid*/
{SO_SENDAUTH, "SO_SENDAUTH"},            /* option number 1*/
{SO_SENDCKSUM, "SO_SENDCKSUM"},          /* option number 2*/
{SO_SENDMD5,   "SO_SENDMD5"},            /* 3*/
{SO_SENDMD5OPT, "SO_SENDMD5OPT"},        /* 4 */
{SO_SENDAHA,    "SO_SENDAHA"},           /* 5 */
{SO_SENDNBOMD5OPT, "SO_SENDNBOMD5OPT"},  /* 6 */
{SO_SENDNULLCKSUM, "SO_SENDNULLCKSUM"},  /* 7 */
{SO_SENDROG, "SO_SENDROG"},              /* 8 */
{SO_SENDNBOROG, "SO_SENDNBOROG"}         /* 9 */
};

int no_auth_options = (sizeof (options)/ sizeof (option)) - 1;


/* AH */

int lsrr = 0;  /* 0 if no source routing is being done */

u_char optlist [MAX_IPOPTLEN];
char * oix;
char n_times; /* when data is being source routed multiple times */
int avndebug = 1;
int avndebug2 = 0;
int get_sroute_options();
int set_auth_option ();

/* source routing end */

/* dotted10
   Checks that a string is a legal (4-part dotted decimal) internet address. 
   Returns the address in network byte order if successful, otherwise, -1.
*/

u_long
dotted10(cp)
   register char *cp;
{
   register u_long val;
   register char c;
   u_long parts[4], *pp = parts;
   int i;

again:
   /* Collect number up to ``.''.  */
   val = 0;
   while ((c = *cp)) {
      if (isdigit(c)) {
         val = (val * 10) + (c - '0');
         cp++;
         continue;
         }
      break;
      }
   if (*cp == '.') {
      /* Internet format: a.b.c.d  */
      if (pp >= parts + 4)
         return (-1);
      *pp++ = val, cp++;
      goto again;
      }
   /* Check for trailing characters. */
   if (*cp && !isspace(*cp))
      return (-1);
   *pp++ = val;
   
   /*  Address must have four parts. */
   if ((pp - parts) != 4)
      return (-1);
      
   /* a.b.c.d -- 8.8.8.8 bits */
   for (i=0;i<4;i++) if (parts[i] > 0xff) return(-1);
   val = (parts[0] << 24) | ((parts[1] & 0xff) << 16) |
            ((parts[2] & 0xff) << 8) | (parts[3] & 0xff);
   val = htonl(val);
   return (val);
} /* dotted10 */

boolean
isbroadcast(ipaddr)
   u_long *ipaddr;
{
   u_long class, host, mask;
   
   class= (*ipaddr >> 28) & 0xd;
   mask = (class == 0)? (0xffffff): 	/* class a */
   	  (class == 0x8)? (0xffff): 	/* class b */
	  (0xff);			/* class c */
   /* *** what about subnet mask? *** */
   host = (*ipaddr & mask);

   return((host==0) || (host==mask)); 
} /* isbroadcast */

/*
*   resolve_name(namep, adrlist, length)
*
*   Resolves host name given in specified server structure, to
*   fill in list of IP addresses and host name.  Returns number of
*   addresses if OK, else zero.
*/
                  
int
resolve_name( namep, adrlist, listlen)
   char *namep;
   u_long adrlist[];
   int listlen;
{
   register int i, count;
   register struct hostent *hp;
   u_long ipaddr;   
   char **ap;
   
   bzero((char *) adrlist, listlen * sizeof(u_long));

  if ((ipaddr = dotted10(namep)) != -1) {
     /* Dotted-decimal... */
     adrlist[0] = ipaddr;
     if (!isbroadcast(&adrlist[0]))
        return(1);
     }
  else	if ((hp = gethostbyname(namep))) {
     count = 0;
     ap = hp->h_addr_list;
     for (i=0;i<listlen && *ap ;i++) {
	count++;
	adrlist[i] = * (u_long *) *ap++;
	if (isbroadcast(&adrlist[i]))
	   adrlist[i--] = 0;
        }
     return(count);
     }

  adrlist[0] = 0;
  return(0);
} /* resolve_name */

/* Error check a string that contains a port number:
	if okay, return(port)
	if bad, return(0) and print to stderr
*/

int
check_port(arg)
   char *arg;
{
   int port;
   char *cp;
	
   for (cp = arg; *cp; cp++) 
      if (! isdigit(*cp))
	 return(-1);
   if (((port = atoi(arg)) < 1024) || (port > 0xffff))
      return(-1);
   return(port);
} /* check_port */

#define END_CODE 0xFF

#define	MAXBUFLEN (256*1024)    /* was 4096, then 32768 */
#define MAXUDPLEN 48000		/* what should this really be? */

#define MAX_RUNS 1000

char *host;
struct in_addr adr;

#define	DEF_BUF_COUNT 2000 		/* was 20000 */
#define	DEF_BUF_SIZE (256*1024) 	/* was 8192 */
#define DEF_UDP_COUNT 20000
#define DEF_UDP_SIZE 8192

#define TCP_WINDOW (48*1024)

#define DEF_TCP_PORT	7000
#define DEF_UDP_PORT	8000

boolean
    tcp = TRUE;
unsigned
    socket_resize = TCP_WINDOW,
    writesPerTick = 0, iterations = 1, 
    lnth = 0, 
    cnt = 0, 
    align = 0xcb0, /* the optimal value is determined experimentally */
    port = 0, 
    elapsed,	/* microseconds */
    send_rate,   /* Kbits/second -- for UDP*/
    rate,       /* Kbits/second */
    times = 0;
double
    sent = 0,
    rcvd = 0;
int
    status_sock, 
    st_listen_sock,
    listen_sock, 
    snd_sock,
    rcv_sock;
struct timeval tv1, tv2;
u_char *buf;
char *buf_base;

extern int errno;
extern char *valloc();

#define IP_ADDR_LEN 4

void
print_ip_addr(data)
     char *data;
{
   int i;

   for (i = 0; i < IP_ADDR_LEN; i++)
      printf("%s%d", (i)?".":" ", (u_char)data[i]);
}

static void
sprint_ip_addr(data, name)
     char *data, *name;
{
   sprintf(name, "%d.%d.%d.%d", (u_char)data[0], 
	   (u_char)data[1], (u_char)data[2], (u_char)data[3]);
}

int
adjust_sock_size(s, which, val)
    int s, which, val;
{
    if (0 > setsockopt(s, SOL_SOCKET, which, &val, sizeof(int))) {
	perror("setsockopt");
	return(FALSE);
	}
    return(TRUE);
} /* adjust_sock_size */

int
set_nodelay(s, val)
    int s, val;
{
   if (0 > setsockopt(s, getprotobyname("tcp")->p_proto,
		      TCP_NODELAY, &val, sizeof(int))) {
      perror("setsockopt");
      return(FALSE);
      }
   return(TRUE);
} /* set_nodelay */

/*************************************************************************/

void
usage()
{
     fprintf(stderr,
            "Usage: blast [-u | -t] [-n] [-c cnt] [-l len] [-p port] [-i times]\
[-s gateway1[:gateway2]  | -S strict_route] host\n%s%s",
            "             [-w writesPerTick]\n",
            "  cnt > 0, len > 0, 0 < times <= 1000; no host required on passive\
 side.\n");
    exit(1);

}

void
net_close()
{
    free(buf_base);

    close(status_sock);
    close(st_listen_sock);

#ifdef ATP
    satp_close(snd_sock);
    satp_close(rcv_sock);
#else
    close(snd_sock);
    close(rcv_sock);
#endif /* ATP */
    close(listen_sock);
    exit(0);
}

int
main(argc, argv)
    char **argv;
{
    unsigned r_rate[MAX_RUNS], s_rate[MAX_RUNS];
    double sd, sd1;
    float average_rate, average_send_rate;
    unsigned
        i, 
	total_Kbps_rcvd = 0, total_Kbps_send = 0, 
	page_size = getpagesize(),
	total_bufs = 0, 
	elapsed_ms, 
	min_rate = 1000000, min_send_rate = 1000000, /* large number */
	max_rate = 0, max_send_rate = 0,
	min_elapsed = 10000000,	/* large number of milliseconds */
        max_elapsed = 0;

        /*source routing begin*/
        /* temp. for determining whether SSRR desired or LSRR */
        char c_option = '\0';
        /* for multiple times iterations */
        n_times = 1;
        /*source routing end*/


    while (argc > 1) {
	if (argv[1][0] == '-') {
	    switch (argv[1][1]) {
		case 't':
		    break;
		case 'u':
		    tcp = FALSE;
		    break;
                /* AH */
                /*Added for command line nagle switching */
	        case 'n':
                    nodelay_on = TRUE;
                    break;

	        case 'h':  /* Was previously 'a't  */
                    if (authenticate)
                    {
                       usage();
                    }
                    if ( argc < 3)
                         usage();
                    authenticate = atoi (argv[2]); /* Set option for AH */
                    if ( (authenticate <= 0) || (authenticate > no_auth_options) )
                         /* If an invalid number (such as 0)
                            is given here it will be caught*/ 
                         usage();
                    /* error check this no. later */
                    argc--;
                    argv++;
                    break;
                /* AH */

		case 'p':
		    if (argc < 3)
			usage();
		    if (!(port = check_port(argv[2])))
			usage();
		    argc--;
		    argv++;
		    break;
		case 'c':
		    if (argc < 3)
			usage();
		    cnt = atoi(argv[2]);
		    if (cnt <= 0)
			usage();
		    argc--;
		    argv++;
		    break;
		case 'a':
		    if (argc < 3)
			usage();
		    align = (int)strtol(argv[2], (char **)NULL, 0);
		    align = (align>>2)<<2; /* we need a word aligned value */
		    if (align < 0 || align > (page_size-1))
			usage();
		    argc--;
		    argv++;
		    break;
		case 'i':
		    if (argc < 3)
			usage();
		    iterations = atoi(argv[2]);
		    if (iterations <= 0 || iterations > MAX_RUNS)
			usage();
		    argc--;
		    argv++;
		    break;
		case 'w':
		    if (argc < 3)
			usage();
		    writesPerTick = atoi(argv[2]);
		    if (writesPerTick <= 0 || iterations > MAX_RUNS)
			usage();
		    argc--;
		    argv++;
		    break;
		case 'l':
		    if (argc < 3)
			usage();
		    lnth = atoi(argv[2]);
		    if (lnth <= 0 || lnth > MAXBUFLEN)
			usage();
		    argc--;
		    argv++;
		    break;
                /* source routing begin*/   
                /* The strict source routing option - SSRR */
		  case 'S':
                     c_option = 'S';
                /* NOTE:  Intentional fall through*/
                /* The loose source routing option */
		  case 's':
                    if (c_option != 'S')
		    {
                        c_option = 's';
		    }
                    if (argc < 3)
		    {
                          usage();
		    }

                    /* In case an attempt is made to specify more than
                        one source route on the command line
                    */

                    if (lsrr)
                    {
                        fprintf(stderr,"\nCan set source route once only\n");
                         usage();
		    }

                    /* get_sroute_options returns a 1 if there is
                       an error in the options
                    */

                     if( get_sroute_options (argv [2], c_option) )
		     {
                       usage();
		     }
                     /* Since a correct source routing option string has
                        been processed the lsrr flag is now set. CHECK
                     */

                     argc--;
                     argv++;
                     break;
                    /* source routing end*/

		default:
		    usage();
	    }
	}
	else {
	    if (host)
		usage();
	    else
		host = argv[1];
	}
	argc--;
	argv++;
    }

    if (!port)
	    port = (tcp) ? DEF_TCP_PORT : DEF_UDP_PORT;

    if (!lnth)
	lnth = (tcp) ? DEF_BUF_SIZE : DEF_UDP_SIZE;
    else
	if (tcp) {
	    if (lnth > MAXBUFLEN)
		usage();
	    }
	else
	    if (lnth > MAXUDPLEN)
                usage();

    if (!cnt)
	cnt = (tcp) ? DEF_BUF_COUNT : DEF_UDP_COUNT;


    /* source routing begin */
    /* Checks to see that the source routing feature has been activated with
    valid other options
    */

    if (lsrr)
    {
       if ( !tcp )
	 {
          fprintf ( stderr, "For now this program implements the source routing\
 option with TCP only\n" );
          usage ();
	}

      if ( !host )
	{
        fprintf ( stderr, "The source routing option can be used from the activ\
e side only. (i.e. the sender only)\n" );
        usage();
        }

    }    
   /* source routing end */
   /* TCP_NODELAY check */
   if (nodelay_on)
   {
      if (!host)
      {
        fprintf( stderr, "TCP_NODELAY command line control for active side only\n");
        usage();
      }
    }
  /* TCP_NODELAY check */


    buf_base = (char *)valloc(MAXBUFLEN+align);
    buf = (u_char *)(buf_base + align);
    printf("alignment = 0x%x; buf = 0x%x\n", align, (unsigned)buf);

/* this allocation hack from netperf doesn't work as well as using
   a fixed allignment offset that has been experimentally determined
   to work the best
    {
    int buf_offset, malloc_size;

    printf("alignment hack: ");
    buf_offset = 0;
    align = 8;
    malloc_size = MAXBUFLEN + align + buf_offset;
    buf_base = (char *)malloc(malloc_size);
    buf = (u_char *)(((long)buf_base + 
		      (long)align - 1) & ~((long)align - 1));
    buf = (u_char *)((char *)buf + buf_offset);
    printf("buf = 0x%x\n", buf);
    }
*/

    if (buf_base == NULL) {
	perror("buf allocation failed\n");
	exit(1);
	}

    if (host == NULL) {
	/* listener */
	signal(SIGINT, net_close);

	if (tcp)
	    tcp_listen();
	else
	    udp_listen();
    }
    else {	/* sender */
	if (! resolve_name(host, (u_long *) & adr.s_addr, 1)) {
	    printf("blast: bad host name\n");
	    usage();
	    }
	else {
	    printf("Sending %d %s bufs of len %d to %s on port %d",
		   cnt, (tcp) ? "TCP" : "UDP", lnth, host, port);

	    if (iterations - 1)
		printf(", %d times...\n", iterations);
	    else
		printf("...\n");
	    fflush(stdout);

	    /* initialize the buffers that will be sent */
	    for (i = 0; i < MAXBUFLEN; i+=2) {
		*(u_short *)&buf[i] = ((u_short)i&0xFFFF)/2;
		if (buf[i] == END_CODE)
		    buf[i] = 0;
		if (buf[i+1] == END_CODE)
                    buf[i+1] = 0;
		}
/*
            for (i = 0; i < MAXBUFLEN; i++)
               buf[i] = 0xFE;
*/
	    if (tcp)
		/* initialize front of buf with the byte count */
		*(double *)buf = (double)lnth * (double)cnt;
	    else
		*(double *)buf = (double)0;

	    for (times = 0; times < iterations; times++) {
		if (iterations - 1)
		    printf("%d\n", times + 1);

		if (tcp)
		    tcp_send();
		else
		    udp_send();
		elapsed_ms = elapsed / 1000;
		total_Kbps_rcvd += rate;
		r_rate[times] = rate;
		total_bufs += (unsigned) (rcvd / lnth);

		if (rate > max_rate) {
		    min_elapsed = elapsed_ms;
		    max_rate = rate;
		}

		if (rate < min_rate) {
		    max_elapsed = elapsed_ms;
		    min_rate = rate;
		}

		if (!tcp) {
		    total_Kbps_send += send_rate;
		    s_rate[times] = send_rate;

		    if (send_rate > max_send_rate)
			max_send_rate = send_rate;

		    if (send_rate < min_send_rate)
			min_send_rate = send_rate;
		    }
		fflush(stdout);
		}
	    if (iterations - 1) {
		printf("\n%d %s bufs of len %d to %s; %d runs; %u bufs rcvd.\n",
		 cnt, (tcp) ? "TCP" : "UDP", lnth, host, times, total_bufs);

		average_rate = (float)((float)total_Kbps_rcvd / 
				       (float)iterations);
		if (!tcp) {
		    average_send_rate = (float)((float)total_Kbps_send / 
						(float)iterations);
		    }
		for (times = 0; times < iterations; times++) {
		    float temp;

		    temp = (float)((float)r_rate[times] - average_rate);
		    temp = temp*temp;
		    sd += (double)temp;
		    if (!tcp) {
			temp = (float)((float)s_rate[times] -
				       average_send_rate);
			temp = temp*temp;
			sd1 += (double)temp;
			}
		    }
		sd = sd / (double)(iterations-1);
		sd = sqrt(sd);
		if (!tcp) {
		    sd1 = sd1 / (double)(iterations-1);
		    sd1 = sqrt(sd1);
		    }

		printf("RCV Rate Mb/s -- Best: %u.%03u, Worst: %u.%03u, Average: %.3f, SD: %.3f\n",
		       max_rate / 1000, max_rate % 1000,
		       min_rate / 1000, min_rate % 1000,
		       average_rate/1000, sd/1000);
		if (!tcp) {
		    printf("SND Rate Mb/s -- Best: %u.%03u, Worst: %u.%03u, Average: %.3f, SD: %.3f\n",
			   max_send_rate / 1000, max_send_rate % 1000,
			   min_send_rate / 1000, min_send_rate % 1000, 
			   average_send_rate/1000, sd1/1000);
		    }
	    }
	    fflush(stdout);
	}
    }	/* sender */

    exit(0);
}			     /* main */


void
tcp_send()
{
    /*source routing begin*/
    struct protoent * pe;   
    /* source routing end */
    double dots = 0, total_bytes = 0;

    struct sockaddr_in to;
    struct timeval timeout;
    int i, j, len, width = getdtablesize();

    if (lnth > MAXBUFLEN) {
	fprintf(stderr, "%d is max buffer size\n", MAXBUFLEN);
	exit(1);
    }

    if ((snd_sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
	perror("tcp socket");
	exit(1);
    }

#ifndef ATP
    if (!adjust_sock_size(snd_sock, SO_RCVBUF, socket_resize)) {
	close(snd_sock);
	exit(1);
    }
    if (!adjust_sock_size(snd_sock, SO_SNDBUF, socket_resize)) {
	close(snd_sock);
	exit(1);
    }
/*  This hurts performance, because smaller packets are sent. */
    /*  Changed  to provide a command-line option */



   if (nodelay_on)
   {
       printf ("Setting TCP_NODELAY to TRUE\n");
       if (! set_nodelay(snd_sock, TRUE)) {
          close(snd_sock);
          exit(1);
       }
   }
#endif /* not ATP */

    bzero(&to, sizeof(to));
    to.sin_family = AF_INET;
    to.sin_port = port;
    to.sin_addr = adr;

#ifdef ATP
    if (0 > satp_connect(snd_sock, &to, sizeof(to))) {
#else
    if (0 > connect(snd_sock, &to, sizeof(to))) {
#endif /* ATP */
	perror("tcp connect");
	exit(1);
    }
    
    
    if ( lsrr)
    {
      if ( n_times){
         
         
         lsrr ++ ;  /* Now we can fill in the total length */
         optlist [IPOPT_OLEN] = IPOPT_MINOFF - 1 + ( lsrr * sizeof (u_long) ) ;
         /* now put the destination as the final entry */
         bcopy ( (caddr_t) &to.sin_addr, oix, sizeof (u_long) );

         oix += sizeof (u_long);

         while ( ((unsigned)oix - (unsigned)optlist) & 3 )
	 {
           oix ++;
         }
         n_times --;
       }

       if ( (pe = getprotobyname ("ip")) == NULL)
       {
          perror ("tcp_send: unknown protocol");
          exit (3);
       }
       if ( avndebug2)
       {

           print_oplist ("FINAL");
       }


       if ( setsockopt (snd_sock, pe->p_proto, IP_OPTIONS, optlist, (int)((unsigned)oix - (unsigned)optlist) ) < 0  )
      {
          perror ("blast: lsrr options" );
          exit (4);
      }else

     {
          if (avndebug2)
	  {
             printf ("Done the setsockopt to insert source routing \n");
	  }

     }

    }

    /*source routing end*/



    /* AH */ 
    if ( authenticate)
    {
       set_auth_option(snd_sock ,authenticate);
    }
   
    /* AH */  

    GET_TIME(tv1);

    /* send all of the data */
    if (!writesPerTick) {
	for (i = cnt; i; i--) {
	    int sent;

#ifdef ATP
	    if (lnth != (sent = satp_send(snd_sock, buf, lnth, 0))) {
		perror("tcp send");
		satp_close(snd_sock);
		exit(1);
		}
#else
	    if (lnth != (sent = send(snd_sock, buf, lnth, 0))) {
		perror("tcp send");
		close(snd_sock);
		exit(1);
		}
#endif /* ATP */
/*
	    else {
		if ((total_bytes+=sent) > dots) {
		    dots+= 10000000;
		    printf(".");
		    fflush(stdout);
		    }
		}
*/
	} /* for */
    }
    else {
	/* as short as possible, really 20 milliseconds */
	timeout.tv_sec = 0;
	timeout.tv_usec = 1;

	for (i = cnt; i;) {
	    /* add delay between packets */
	    (void) select(width, NULL, NULL, NULL, &timeout);

	    for (j = writesPerTick; j && i; j--, i--)
		{
#ifdef ATP
		    if (0 > satp_send(snd_sock, buf, lnth, 0)) {
			perror("tcp send");
			satp_close(snd_sock);
			exit(1);
			}
#else
		    if (0 > send(snd_sock, buf, lnth, 0)) {
			perror("tcp send");
			close(snd_sock);
			exit(1);
			}
#endif /* ATP */
		}
	    } /* for */
    }

    /* get the count sent by the receiver */
#ifdef ATP
    if (0 > (len = satp_recv(snd_sock, buf, sizeof(double), 0))) {
	satp_close(snd_sock);
	perror("tcp recv");
	exit(1);
	}
#else
    if (0 > (len = recv(snd_sock, buf, sizeof(double), 0))) {
	close(snd_sock);
	perror("tcp recv");
	exit(1);
	}
#endif ATP
    rcvd = *(double *) buf;

    GET_TIME(tv2);
    SUB_TIME(tv2, tv1);

#ifdef ATP
    satp_close(snd_sock);
#else
    close(snd_sock);
#endif /* ATP */

    /* convert to microseconds */
    elapsed = tv2.tv_sec * 1000000 + tv2.tv_usec;
    rate = (unsigned) KBITS_PER_SEC(rcvd, elapsed);

    printf("\tTime: %u.%03u s, Rate: %u.%03u Mb/s\n",
	   elapsed / 1000000, (elapsed % 1000000) / 1000,
	   rate / 1000, rate % 1000);
} /* tcp_send */

void
tcp_listen()
{
    struct sockaddr_in to, remote;
    int i, j, socklen, len, val;
    double max_byte_count;

    if ((listen_sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
	perror("tcp socket");
	exit(1);
    }

    bzero(&to, sizeof(to));
    to.sin_family = AF_INET;
    to.sin_port = port;
    to.sin_addr.s_addr = INADDR_ANY;

    if (0 > bind(listen_sock, &to, sizeof(to))) {
	perror("tcp bind");
	exit(1);
    }

#ifndef ATP
    if (!adjust_sock_size(listen_sock, SO_RCVBUF, socket_resize,
			  "SO_RCVBUF")) {
	close(listen_sock);
	exit(1);
	}
    if (!adjust_sock_size(listen_sock, SO_SNDBUF, socket_resize,
			  "SO_SNDBUF")) {
	close(listen_sock);
	exit(1);
	}
    if (!set_nodelay(listen_sock, TRUE)) {
	close(listen_sock);
	exit(1);
	}
#endif /* ATP */

    while (TRUE) {
	/* for now, just handle one connection at a time */
	if (0 > listen(listen_sock, 1)) {
	    perror("tcp listen");
	    close(listen_sock);
	    exit(1);
	    }
	printf("Listening on TCP port %d ...\n", port);

	socklen = sizeof(remote);
#ifdef ATP
	if (0 > (rcv_sock = satp_accept(listen_sock, &remote, &socklen))) {
	    perror("tcp accept");
	    satp_close(rcv_sock);
	    exit(1);
	    }
#else
	if (0 > (rcv_sock = accept(listen_sock, &remote, &socklen))) {
	    perror("tcp accept");
	    close(rcv_sock);
	    exit(1);
	    }
#endif /* ATP */

	printf("Receiving TCP data from");
	print_ip_addr((char *) &remote.sin_addr.s_addr);
	printf(" ... \n");

	{
	rcvd = 0;
	max_byte_count = MAXBUFLEN; /* make gcc happy */
	do {

#ifdef ATP
	    if (0 > (len = satp_recv(rcv_sock, buf, MAXBUFLEN, 0))) {
		perror("tcp recv");
		satp_close(rcv_sock);
		exit(1);
		}
#else
	    if (0 > (len = recv(rcv_sock, buf, MAXBUFLEN, 0))) {
		perror("tcp recv");
		close(rcv_sock);
		exit(1);
		}
#endif /* ATP */
	    if (rcvd == 0)
		max_byte_count = *(double *)buf;
	    rcvd += len;
	    } while (rcvd < max_byte_count);
	*(double *) buf = rcvd;

#ifdef ATP
	if (0 > satp_send(rcv_sock, buf, sizeof(double), 0)) {
	    perror("tcp send");
	    satp_close(rcv_sock);
	    exit(1);
	    }
#else
	if (0 > send(rcv_sock, buf, sizeof(double), 0)) {
	    perror("tcp send");
	    close(rcv_sock);
	    exit(1);
	    }
#endif /* ATP */
	}

#ifdef ATP
	satp_close(rcv_sock);
#else
	close(rcv_sock);
#endif /* ATP */

	} /* while TRUE */
} /* tcp_listen */

void
udp_send()
{
    struct sockaddr_in remote, me;
    struct timeval timeout;
    int width = getdtablesize();
    int i, j, len;
    u_char lastbuf[1];

    for (i = 0; i < sizeof(lastbuf); i++)
	lastbuf[i] = END_CODE;

    {
    struct sockaddr_in to, remote;
    int i, j, socklen, len, val;

    if ((status_sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
	perror("tcp socket (status)");
	exit(1);
	}

    bzero(&to, sizeof(to));
    to.sin_family = AF_INET;
    to.sin_port = port+1;
    to.sin_addr = adr;

    if (0 > connect(status_sock, &to, sizeof(to))) {
	perror("tcp connect");
	exit(1);
	}
    sleep(1);
    }

    if ((snd_sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
	perror("udp socket");
	exit(1);
	}

    if (!adjust_sock_size(snd_sock, SO_SNDBUF, socket_resize,
			  "SO_SNDBUF")) {
	close(snd_sock);
	exit(1);
	}

    if (lnth > MAXBUFLEN) {
	fprintf(stderr, "%d is max buffer size\n", MAXBUFLEN);
	close(snd_sock);
	exit(1);
	}

    bzero(&remote, sizeof(remote));
    remote.sin_family = AF_INET;
    remote.sin_port = port;
    remote.sin_addr = adr;

    if (0 > connect(snd_sock, &remote, sizeof(remote))) {
	perror("udp connect");
	close(snd_sock);
	exit(1);
	}

    /* AH */
   

    if (authenticate)
    {
      set_auth_option(snd_sock,authenticate);
    } 

    GET_TIME(tv1);

    /* send all of the data */
    if (!writesPerTick) {
	for (i = cnt; i; i--) {
	    if (0 > send(snd_sock, buf, lnth, 0)) {
		perror("udp send");
		exit(1);
		}
	    }
	}
    else {
	/* as short as possible, really 20 milliseconds */
	timeout.tv_sec = 0;
	timeout.tv_usec = 1;

	for (i = cnt; i;) {
	    /* add delay between packets */
	    (void) select(width, NULL, NULL, NULL, &timeout);

	    for (j = writesPerTick; j && i; j--, i--)
		if (0 > send(snd_sock, buf, lnth, 0)) {
		    perror("udp send");
		    exit(1);
		    }
	    }
	}

    /* The end-pattern packet might be lost.  Do we need to send a bunch? */

    /* send end-message on the status connection, too */
    if (0 > send(status_sock, lastbuf, sizeof(lastbuf), 0)) {
        perror("tcp send");
        exit(1);
        }

    /* get the count sent by the receiver on the status connection */
    if (0 > (len = recv(status_sock, buf, sizeof(double), 0))) {
	    perror("status recv");
	    exit(1);
	    }
	rcvd = *(double *) buf;

    GET_TIME(tv2);
    SUB_TIME(tv2, tv1);

    close(snd_sock);
    close(status_sock);

    printf("\tRemote user process: %.0f buffers received.\n", rcvd);

    /* convert to bytes */
    rcvd = rcvd * (double)lnth;
    sent = (double)cnt * (double)lnth;

    /* convert to microseconds */
    elapsed = tv2.tv_sec * 1000000 + tv2.tv_usec;
    rate = (unsigned) KBITS_PER_SEC(rcvd, elapsed);
    send_rate = (unsigned) KBITS_PER_SEC(sent, elapsed);

    printf("\tTime: %u.%03u s, RCV Rate: %u.%03u, SND Rate: %u.%03u Mb/s\n",
	   elapsed / 1000000, (elapsed % 1000000) / 1000,
	   rate / 1000, rate % 1000, send_rate / 1000, send_rate % 1000);
/*
   printf("\tTime: %.3f seconds, Rate: %.3f Mb/sec\n", elapsed/1000.0,
	  (lnth*rcvd)/((elapsed *1000.0)/8));
*/
} /* udp_send */

void
udp_listen()
{
    struct sockaddr_in to, remote;
    int i, j, 
        socklen = sizeof(remote), 
        val, len, 
        width = getdtablesize();
    fd_set readfds;

    if ((st_listen_sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
	perror("tcp socket");
	exit(1);
	}

    bzero(&to, sizeof(to));
    to.sin_family = AF_INET;
    to.sin_port = port+1;
    to.sin_addr.s_addr = INADDR_ANY;

    if (0 > bind(st_listen_sock, &to, sizeof(to))) {
	perror("tcp bind (status)");
	exit(1);
	}

    if ((rcv_sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
	perror("udp socket");
	exit(1);
	}

    if (!adjust_sock_size(rcv_sock, SO_RCVBUF, socket_resize,
			  "SO_RCVBUF")) {
	close(rcv_sock);
	exit(1);
	}

    bzero(&remote, sizeof(remote));
    remote.sin_family = AF_INET;
    remote.sin_port = port;
    remote.sin_addr.s_addr = INADDR_ANY;

    if (0 > bind(rcv_sock, &remote, sizeof(remote))) {
	perror("udp bind");
	exit(1);
	}

    printf("Receiving on UDP port %d ...\n", port);

    for (i = 0;; i++) {
        if (0 > listen(st_listen_sock, 1)) {
            perror("tcp listen (status)");
            exit(1);
            }
	printf("Listening on TCP port %d ...\n", port+1);

        socklen = sizeof(remote);
        if (0 > (status_sock = accept(st_listen_sock, &remote, &socklen))) {
            perror("tcp accept (status)");
            exit(1);
            }

	if (fcntl(status_sock, F_SETFL, FNDELAY) < 0){
	    perror("fcntl F_SETFL, FNDELAY");
	    net_close();
	    exit(1);
	    } 

        printf("Receiving data from");
        print_ip_addr((char *) &remote.sin_addr.s_addr);
        printf(" ... \n");
        fflush(stdout);

	rcvd = 0;
	FD_ZERO(&readfds);
	FD_SET(status_sock, &readfds);
	FD_SET(rcv_sock, &readfds);
	while (TRUE) {
	    select(width, readfds, NULL, NULL, NULL);

	    if (FD_ISSET(status_sock, &readfds))
		if (0 < (len = recv(status_sock, buf, sizeof(double), 0))) {
		    if (buf[len-1] == END_CODE) {
			printf("\tReceived %.0f buffers\n", rcvd);
			break;
			}
		    }

	    if (FD_ISSET(rcv_sock, &readfds))
		if (0 > (len = recvfrom(rcv_sock, buf, MAXBUFLEN, 0, &remote, 
					&socklen))) {
		    perror("udp recvfrom");
		    net_close();
		    exit(1);
		    }
	        else
		    rcvd++;
	    }

	/* send the count of buffers received to the sender */
	*(double *) buf = rcvd;
	if (0 > send(status_sock, buf, sizeof(double), 0)) {
	    perror("sending count");
	    net_close();
	    exit(1);
	    }

	close(status_sock);

	} /* for */
} /* udp_listen */

/* source routing begin*/

/************************************************************************
Function :       get_sroute_options()
Input    :  a pointer to a char which is the name of the gateways
            to traverse in the route and a char indicating loose
            source route or strict source route
Output   :  0 if the inp_str was correctly interpreted.
            1 if there was an error in interpretation.

Comments :  This function checks that
           - The input string consists of valid hostnames or IP addresses
             separated by ":"
           - The no. of intermediate gateways does not exceed those allowd
             by ip options
           On successful exit the ip source routing option is setup
           in the GLOBAL optlist.


************************************************************************/
int
get_sroute_options ( inp_str, c_opt )
char * inp_str;
char  c_opt;
{
      char separator[2];  /* strtok() looks like a good function */
      char * gateway;
      strcpy (separator,":");


      oix = (char *)  optlist;
      bzero ( optlist, sizeof (optlist) );
      *oix++ = (c_opt == 'S')? IPOPT_SSRR : IPOPT_LSRR;
      *oix++ ;  /* Will fill in the total length of option later */
      *oix++ = IPOPT_MINOFF ;  /* Pointer to the source routing addresses */


      if ( (gateway = strtok (inp_str, separator) ) != NULL)
	{
         /* Check to see if the gateway is a valid host name or
            a valid host addr
         */
         if ( add_if_ok (gateway) )
	   {
            goto bad;
	  }
         while ( ( gateway = strtok ( (char*)NULL, separator)) )
	   {

           if ( add_if_ok (gateway) )
	     {
              goto bad;
	    }
         }
         return (0);      /* Everything is ok */

       }else
	 {
	   bad :    return (1);   /* a problem somewhere */

	 }


    }

int
add_if_ok ( gateway )
char * gateway ;
{
     struct in_addr gate_adr;

     if (resolve_name (gateway, (u_long*) &gate_adr.s_addr, 1) )
       {
        if( (lsrr + 1) >= ( (MAX_IPOPTLEN - IPOPT_MINOFF) / sizeof(u_long) ) )
	  {
           fprintf (stderr, "No more than %d gateways\n",
              ( (MAX_IPOPTLEN - IPOPT_MINOFF) / sizeof(u_long)) - 1 );
            return (1);
	 }

        lsrr ++;  /* Added a gateway to the optlist */

       /* Since valid put it on the options list */
        bcopy ( &gate_adr.s_addr, oix, sizeof (u_long) );
        if (avndebug)
	  {
           printf ("Source route through  \"%s\" having address   ", gateway);
           print_ip_addr (oix);
           printf ("\n");

	 }
      oix += sizeof (u_long); /* Now points to the next location in the optio\
n list */
       return (0);
      }else
	{
        if (avndebug)
	  {
           printf ("Unable to resolve address for gateway %s\n", gateway);
	 }

           return (1);

      }

   }
 

int
print_oplist ( string )
char * string;

{

    int counter = 1;
    char * pointer;
    pointer = (char *) optlist;

    printf ("Contents of the oplist array at the %s stage \n", string);

    printf (" Starting address is %x \n", (int)&optlist[0] );

    printf (" Value of option = %d \n", (int)optlist[0] ) ;

    printf ("Length of the option is %d bytes\n", (int)optlist[1]);

    printf ("Offset of the options is %d \n" , (int) optlist[2] );

    printf ("oix - oplist is %u \n " , (unsigned)oix - (unsigned)&optlist[0] );

    printf ("The addresses are\n");

    pointer = (char *) &optlist[IPOPT_MINOFF - 1];

    for ( counter = 1; counter <= lsrr; counter ++)
      {
        print_ip_addr ( pointer);
        printf ("\n");
        pointer  += 4;
      }

  }


/* source routing end */

/* AH */
int set_auth_option ( s, auth )
int s;
int auth;
{
    	int setval, getval;
        int size = sizeof (int);
        u_int option_value;        
        char error_string[50];
   
        setval = 1;
        getval = 0;
	printf ("\nDoing a setsockopt with  ");
        val_print ( setval, getval);
        

       /* assumes that a sanity check has been done on the 
	  values passed in as auth
       */
               
        printf ("Setting the %s option\n", options[auth].option_name);
        option_value = options[auth].option_value;
	printf ("Value of %s is %x\n", options[auth].option_name,option_value); 
             
	if (setsockopt ( s, SOL_SOCKET,option_value, (char *) &setval, sizeof (int)) < 0)
	{
	  sprintf (error_string,"blast:setsockopt for %s",
		   options[auth].option_name);
	  perror (error_string);
	  exit (101);
	}
       
	printf ("Doing a getsockopt() with    ");
        val_print (setval, getval);
 

	if (getsockopt ( s, SOL_SOCKET,option_value, (char *) &getval, &size) < 0 )
	  { 
	    sprintf(error_string, "blast:getsockopt for %s",
		    options[auth].option_name);
	    /*  printf("\nValue of option_value is %d\n", option_value);*/
	    perror(error_string);
	    exit(911);
	  }


        printf("Values after getsockopt() ");  
        val_print ( setval, getval);
      
}




int val_print ( setval, getval)
int setval;
int getval;

{
   printf ("setval = 0x%x, getval = 0x%x\n", setval, getval);
 
 }        
     

/* AH */
